# Author: Acer Zhang
# Datetime:2021/7/7 
# Copyright belongs to the author.
# Please indicate the source for reprinting.
import traceback
from qpt.kernel.qos import Logging

SUPPORT_AVX = None
try:
    from paddle.fluid.core_avx import *
    SUPPORT_AVX = True
except Exception as e:
    Logging.warning(str(e))
    info = traceback.format_exc()
    if "cv2" in info:
        # ToDo 实际上还是不能确定，但Paddle这块不知道为什么要这样测cv2是否存在
        SUPPORT_AVX = True
    else:
        SUPPORT_AVX = False



# Author: Acer Zhang
# Datetime: 2021/5/26
# Copyright belongs to the author.
# Please indicate the source for reprinting.

import os
import pickle
import datetime
from distutils.sysconfig import get_python_lib

from qpt.kernel.qlog import Logging
from qpt.memory import QPT_MODE, CheckRun

# 定义优先级 优先级越高执行顺序越考前，一般设置为GENERAL_LEVEL
TOP_LEVEL = 5.  # 底层高优先级
TOP_LEVEL_REDUCE = 4.5  # 系统级高优先级
HIGH_LEVEL = 4.  # 高优先级
HIGH_LEVEL_REDUCE = 3.5  # 紧随高优先级
GENERAL_LEVEL = 3.  # 普通优先级
GENERAL_LEVEL_REDUCE = 2.5  # 紧随普通优先级
LOW_LEVEL = 2.  # 低优先级 - BatchInstallation
LOW_LEVEL_REDUCE = 1.5  # 紧随低优先级 - PaddlePaddleCheckAVX
BOTTOM_LEVEL = 1.  # 系统级低优先级
BOTTOM_LEVEL_REDUCE = 0.5  # 底层低优先级


class SubModuleOpt:
    """
    自定义子模块操作，用于子模块封装时和封装后的操作流程设置，支持shell操作和Python原生语言操作
    """

    def __init__(self, disposable=False):
        self.name = self.__class__.__name__

        # 算子是否为一次性算子 - 通常用于安装第三方库等只需要进行一次就可以永久使用的情况
        self.disposable = disposable

        # 环境变量
        # 解释器所在路径占位
        self._interpreter_path = "./"
        # Module目录 - 创建Module时的保存目录/执行Module时的Module目录
        self._module_path = "./"
        # 终端占位
        self._terminal = None
        # 工作目录占位 - 创建Module时的工作目录（待打包的目录）/执行Module时的resources目录
        self._work_dir = "./"

    def act(self) -> None:
        """
        使用Python语句和终端来执行操作
        Example:
            class MyOpt(SubModuleOpt):
                def run_py(self):
                    super().run_py()
                    # 例如新建C:/abc目录
                    import os
                    os.mkdir(r"C:/abc")

                    # 例如在终端中查看当前目录（Windows为dir命令）
                    self.terminal("dir")

                    # 例如在用户使用时为其Module所在的目录中新建abc目录
                    import os
                    os.mkdir(os.path.join(self.module_path, "abc"))
        """
        pass

    def run(self, op_path):
        inactive_file = op_path + ".inactive"
        if (self.disposable and os.path.exists(inactive_file)) and CheckRun.check_run_file(self.config_path):
            Logging.debug(f"找到该OP状态文件{self.name}.inactive，故跳过该OP")
        else:
            self.act()
        if self.disposable and os.path.exists(os.path.dirname(op_path)):
            with open(inactive_file, "w", encoding="utf-8") as f:
                f.write(f"于{str(datetime.datetime.now())}创建了该状态文件")

    @property
    def interpreter_path(self):
        return self._interpreter_path

    @property
    def module_path(self):
        return self._module_path

    @property
    def work_dir(self):
        return self._work_dir

    @property
    def config_path(self):
        return os.path.join(self.module_path, "configs")

    @property
    def site_package_path(self):
        sp_path = os.path.abspath(get_python_lib())
        return sp_path

    @property
    def module_site_package_path(self):
        sp_path = os.path.abspath(os.path.join(self.module_path, "Python/Lib/site-packages"))
        return sp_path

    def prepare(self, work_dir=None, interpreter_path=None, module_path=None, terminal=None):
        self._work_dir = work_dir
        self._interpreter_path = interpreter_path
        self._module_path = module_path
        self._terminal = terminal

    def terminal(self, shell):
        self._terminal(shell)


class SubModule:
    def __init__(self, name=None, level: int = GENERAL_LEVEL):
        if name is None:
            name = self.__class__.__name__
        self.name = name
        self.level = level

        # 占位OP
        self.pack_opts = list()
        self.unpack_opts = list()
        self.ready_unpack_opt_count = 0
        self.details = {"Pack": [], "Unpack": []}
        self._ext_module = list()

        # 占位out_dir，将会保存序列化文件到该目录，pack时需要被set
        self._module_path = "./"
        self._interpreter_path = "./"
        self._terminal = "./"
        self._work_dir = "./"

    def add_ext_module(self, module):
        """
        额外的Module
        :param module:额外的Module
        """
        # ToDo 加个raise，避免没有完成__init__就开始add，针对super__init__的情况
        Logging.info(f"{self.__class__.__name__}中自动添加了名为{module.name}的ExtModule")
        self._ext_module.append(module)

    def get_all_module(self) -> list:
        return [self] + self._ext_module

    def prepare(self, work_dir=None, interpreter_path=None, module_path=None, terminal=None):
        self._work_dir = work_dir
        self._interpreter_path = interpreter_path
        self._module_path = module_path
        self._terminal = terminal

    def add_pack_opt(self, opt: SubModuleOpt):
        self.details["Pack"].append(opt.__class__.__name__)
        self.pack_opts.append(opt)

    def add_unpack_opt(self, opt: SubModuleOpt):
        self.details["Unpack"].append(opt.__class__.__name__)
        self.unpack_opts.append(opt)

    def pack(self):
        """
        在撰写该Module时，开发侧需要的操作
        """
        assert self._module_path, "SubModule的out_dir未设置！"
        for opt in self.pack_opts:
            Logging.info(f"正在加载{self.name}-{opt.name}OP")
            op_path = os.path.join(self._module_path, "opt", self.name, opt.name)
            opt.prepare(interpreter_path=self._interpreter_path,
                        module_path=self._module_path,
                        terminal=self._terminal,
                        work_dir=self._work_dir)
            opt.run(op_path)

        for opt in self.unpack_opts:
            Logging.info(f"正在封装{self.name}-{opt.name}OP")
            self._serialize_op(opt)

    def unpack(self):
        """
        用户使用该Module时，需要完成的操作
        """
        files = os.listdir(os.path.join(self._module_path, "opt", self.name))
        ops = list()
        for file in files:
            if file[:3].isdigit():
                ops.append(file)
            else:
                Logging.warning(f"{self.name}-{file} op可能已损坏，建议重新解压并运行或联系软件提供者")
        ops.sort(key=lambda x: int(x[:3]))
        for op_name in ops:
            op_name = str(op_name)
            if os.path.splitext(op_name)[-1] == ".op":
                op_path = os.path.join(self._module_path, "opt", self.name, op_name)
                with open(op_path, "rb") as file:
                    opt = pickle.load(file)
                    opt.prepare(interpreter_path=self._interpreter_path,
                                module_path=self._module_path,
                                terminal=self._terminal,
                                work_dir=self._work_dir)
                    if QPT_MODE != "Run":
                        Logging.debug(f"正在加载{self.name}-{opt.name}OP")
                    opt.run(op_path)

    # ToDo:做序列化来保存
    def _serialize_op(self, opt):
        name = opt.__class__.__name__
        self.ready_unpack_opt_count += 1
        serialize_path = os.path.join(self._module_path, "opt", self.name)
        serialize_file_path = os.path.join(serialize_path, f"{self.ready_unpack_opt_count:03d}-{name}.op")

        os.makedirs(serialize_path, exist_ok=True)

        with open(serialize_file_path, "wb") as file:
            pickle.dump(opt, file)



import os
import sys
import zipfile

from qpt.modules.base import SubModule, SubModuleOpt, TOP_LEVEL
from qpt.kernel.qlog import Logging
from qpt.kernel.qos import download, get_qpt_tmp_path, copytree
from qpt.memory import QPT_MEMORY

"""
Python镜像打包指南
1. 下载嵌入式版本Python
2. 删除__pycache__
3. ADD 运行库
4. 使用LZMA-64-64 zip的格式上传至文件床，必要时设置密码
"""

PYTHON_ENV_MODE_SPEED_FIRST = "预先解压好Python环境，占用部分硬盘资源但能减少用户使用时速度损失"
PYTHON_ENV_MODE_PACKAGE_VOLUME_FIRST = "[暂不支持]封装后保留压缩的Python环境以减少硬盘资源占用"
PYTHON_ENV_MODE_ONLINE_INSTALLATION = "[暂不支持]不封装Python环境，用户使用时在线进行下载并部署"
DEFINE_PYTHON_ENV_MODE = PYTHON_ENV_MODE_SPEED_FIRST

RESOURCES_URLS = {"Python3.7Env-Win": "https://bj.bcebos.com/v1/ai-studio-online/470b6a91063f425db72ddece0931bcc56"
                                      "6950d67f3914339b40beea878461f3b?responseContentDisposition=attachment%3B%20"
                                      "filename%3DPython37.zip",
                  "Python3.8Env-Win": "https://bj.bcebos.com/v1/ai-studio-online/d0383fa2194d4853883f89b3ab54e841f"
                                      "8cdcc3c88824e2398ba22f9b8fd5094?responseContentDisposition=attachment%3B%20"
                                      "filename%3DPython38.zip",
                  "Python3.9Env-Win": "https://bj.bcebos.com/v1/ai-studio-online/ea6af2820f594d209d6dc421b9392327d"
                                      "270368a6fb2428ebf8bd90cd74d7e26?responseContentDisposition=attachment%3B%20"
                                      "filename%3DPython39.zip",
                  "Python3.10Env-Win": "https://bj.bcebos.com/v1/ai-studio-online/bad0502d69344493a12a10fac443bad"
                                       "ca01a97001c294009937964a6ebb59893?responseContentDisposition=attachment%3B%20"
                                       "filename%3DPython310.zip",
                  "DEFAULT-Win": "https://bj.bcebos.com/v1/ai-studio-online/d0383fa2194d4853883f89b3ab54e841f"
                                 "8cdcc3c88824e2398ba22f9b8fd5094?responseContentDisposition=attachment%3B%20"
                                 "filename%3DPython38.zip"}

DEFAULT_PYTHON_IMAGE_VERSION = "3.8"


class PackPythonEnvOpt(SubModuleOpt):
    def __init__(self, url: str = None, mode=PYTHON_ENV_MODE_SPEED_FIRST):
        super().__init__()
        self.url = url
        self.mode = mode

    def act(self) -> None:
        if self.mode == PYTHON_ENV_MODE_SPEED_FIRST:
            dir_name = get_qpt_tmp_path(os.path.join("Python", "".join(list(filter(str.isdigit, self.url)))[-10:]))
            cache_path = get_qpt_tmp_path(
                os.path.join("Python", "".join(list(filter(str.isdigit, self.url)))[-10:]
                             , "unzip"))
            Logging.info(f"正在加载Python解释器原文件至{dir_name}")
            d_result = download(self.url, "Python.zip", dir_name)
            if d_result:
                zip_path = os.path.join(dir_name, "Python.zip")
                # 解压至输出文件夹
                with zipfile.ZipFile(zip_path) as zip_obj:
                    zip_obj.extractall(cache_path, pwd="gt_qpt".encode("utf-8"))
            copytree(cache_path, os.path.join(self.module_path, "Python"))

        # ToDo 后续支持
        elif self.mode == PYTHON_ENV_MODE_PACKAGE_VOLUME_FIRST:
            Logging.info(f"正在加载Python解释器原文件")
            download(self.url, "Python.zip", self.module_path)
        elif self.mode == PYTHON_ENV_MODE_ONLINE_INSTALLATION:
            pass


class UnPackPythonEnvOpt(SubModuleOpt):
    def __init__(self, url: str = None, mode=PYTHON_ENV_MODE_SPEED_FIRST):
        super().__init__()
        self.url = url
        self.mode = mode

    def act(self) -> None:
        if self.mode == PYTHON_ENV_MODE_SPEED_FIRST:
            pass
        elif self.mode == PYTHON_ENV_MODE_PACKAGE_VOLUME_FIRST:
            zip_path = os.path.join(self.module_path, "Python.zip")
            # 解压至输出文件夹
            with zipfile.ZipFile(zip_path) as zip_obj:
                zip_obj.extractall(os.path.join(self.module_path, "Python"), pwd="gt_qpt".encode("utf-8"))
        elif self.mode == PYTHON_ENV_MODE_ONLINE_INSTALLATION:
            dir_name = get_qpt_tmp_path(os.path.join("Python", "".join(list(filter(str.isdigit, self.url)))[-10:]))
            Logging.debug(f"正在下载Python解释器原文件至{dir_name}")
            download(self.url, "Python.zip", dir_name)
            zip_path = os.path.join(dir_name, "Python.zip")
            # 解压至输出文件夹
            with zipfile.ZipFile(zip_path) as zip_obj:
                zip_obj.extractall(os.path.join(self.module_path, "Python"), pwd="gt_qpt".encode("utf-8"))

        # 添加Python以及Python/lib/python/site_packages_path下的包到环境变量/工作目录
        python_path = self.interpreter_path
        site_packages_path = os.path.join(self.interpreter_path, QPT_MEMORY.site_packages_path)
        script_path = os.path.join(self.interpreter_path, "Scripts")
        sys.path.append(python_path)
        sys.path.append(site_packages_path)
        sys.path.append(script_path)


class BasePythonEnv(SubModule):
    def __init__(self, name, version=None, mode=DEFINE_PYTHON_ENV_MODE, url=None):
        if version:
            resources_name = f"Python{version}Env-Win"
            if resources_name in RESOURCES_URLS:
                Logging.info(f"已在QPT中找到{resources_name}镜像")
            else:
                Logging.warning(f"----------------------------------------------------------------------\n"
                                f"未在QPT中找到{resources_name}镜像，QPT目前提供的Python镜像版本有限，\n"
                                f"请尽可能使用Python3.7/Python3.8/Python3.9等主流Python版本进行打包，兼容性如下。\n"
                                f"----------------------------------------------------------------------\n"
                                f"Python版本|XP Win7 Win8.1 Win10+\n"
                                f"Python37 | X   1     1      1  \n"
                                f"Python38 | X   1     -      1  \n"
                                f"Python39 | X   X     -      1  \n"
                                f"----------------------------------------------------------------------\n"
                                f"已强制设置目标版本号为{DEFAULT_PYTHON_IMAGE_VERSION}，请检查存在以下兼容性问题：\n"
                                f"1. 需考虑待打包的代码是否兼容该Python版本\n"
                                f"2. 该Python版本是否可以在目标用户操作系统上执行\n"
                                f"----------------------------------------------------------------------\n")
                resources_name = "DEFAULT-Win"
            url = RESOURCES_URLS[resources_name]
        assert url and version, "请至少设置url和version中任一字段"
        super().__init__(name, level=TOP_LEVEL)
        self.add_pack_opt(PackPythonEnvOpt(url=url, mode=mode))
        self.add_unpack_opt(UnPackPythonEnvOpt(url=url, mode=mode))
        self.python_version = "非标准的PythonSubModule，需指定版本号"


class AutoPythonEnv(BasePythonEnv):
    def __init__(self, mode=DEFINE_PYTHON_ENV_MODE):
        import platform
        version = platform.python_version()
        Logging.info(f"当前解释器版本为{version}，正在向QPT查询是否存在合适的Python镜像...")
        # 截断版本号，只保留两位
        version = "".join([v if version_index == 1 else v + "."
                           for version_index, v in enumerate(version.split(".")[:2])])
        super().__init__(name=None, url=None, mode=mode, version=version)


class Python37(BasePythonEnv):
    def __init__(self, mode=DEFINE_PYTHON_ENV_MODE):
        super().__init__(name=None, version="3.7", mode=mode)


class Python38(BasePythonEnv):
    def __init__(self, mode=DEFINE_PYTHON_ENV_MODE):
        super().__init__(name=None, version="3.8", mode=mode)


class Python39(BasePythonEnv):
    def __init__(self, mode=DEFINE_PYTHON_ENV_MODE):
        super().__init__(name=None, version="3.9", mode=mode)


class Python310(BasePythonEnv):
    def __init__(self, mode=DEFINE_PYTHON_ENV_MODE):
        super().__init__(name=None, version="3.9", mode=mode)



# Author: Acer Zhang
# Datetime: 2021/5/26 
# Copyright belongs to the author.
# Please indicate the source for reprinting.

import os

from qpt.version import version as qpt_version
from qpt.modules.base import SubModule, SubModuleOpt, TOP_LEVEL_REDUCE, LOW_LEVEL
from qpt.kernel.qos import FileSerialize
from qpt.kernel.qlog import Logging
from qpt.kernel.qcode import PythonPackages
from qpt.memory import QPT_MEMORY

DOWN_PACKAGES_RELATIVE_PATH = "opt/packages"

# 第三方库部署方式
LOCAL_DOWNLOAD_DEPLOY_MODE = "为用户准备Whl包，首次启动时会自动安装，即使这样也可能会有兼容性问题"
LOCAL_INSTALL_DEPLOY_MODE = "[不推荐]预编译第三方库，首次启动无需安装但将额外消耗硬盘空间，可能会有兼容性问题并且只支持二进制包"
ONLINE_DEPLOY_MODE = "用户使用时在线安装Python第三方库"
DEFAULT_DEPLOY_MODE = LOCAL_DOWNLOAD_DEPLOY_MODE

# 第三方库下载版本
PACKAGE_FOR_PYTHON38_VERSION = "3.8"
DEFAULT_PACKAGE_FOR_PYTHON_VERSION = None  # None表示不设置


def set_default_deploy_mode(mode):
    """
    设置全局部署方式
    :param mode: 部署方式
    """
    global DEFAULT_DEPLOY_MODE
    DEFAULT_DEPLOY_MODE = mode


def set_default_package_for_python_version(version):
    """
    设置全局下载的Python包默认解释器版本号
    :param version: Python版本号
    """
    global DEFAULT_PACKAGE_FOR_PYTHON_VERSION
    DEFAULT_PACKAGE_FOR_PYTHON_VERSION = version


class DownloadWhlOpt(SubModuleOpt):
    def __init__(self,
                 package: str,
                 version: str = None,
                 no_dependent=False,
                 find_links: str = None,
                 python_version=DEFAULT_PACKAGE_FOR_PYTHON_VERSION,
                 opts: str = None):
        super().__init__()
        self.package = package
        self.no_dependent = no_dependent
        self.find_links = find_links
        self.opts = opts
        self.version = version
        self.python_version = python_version

    def act(self) -> None:
        if "[FLAG-FileSerialize]" in self.package[:32]:
            self.package = self.package.strip("[FLAG-FileSerialize]")
            self.package = "-r " + FileSerialize.serialize2file(self.package)
        QPT_MEMORY.pip_tool.download_package(self.package,
                                             version=self.version,
                                             save_path=os.path.join(self.module_path, DOWN_PACKAGES_RELATIVE_PATH),
                                             no_dependent=self.no_dependent,
                                             find_links=self.find_links,
                                             python_version=self.python_version,
                                             opts=self.opts)


class LocalInstallWhlOpt(SubModuleOpt):
    def __init__(self,
                 package: str,
                 version: str = None,
                 no_dependent=False,
                 opts: str = None):
        super().__init__(disposable=True)
        self.package = package
        self.no_dependent = no_dependent
        self.opts = opts
        self.version = version

    def act(self) -> None:
        if "[FLAG-FileSerialize]" in self.package[:32]:
            self.package = self.package.strip("[FLAG-FileSerialize]")
            self.package = "-r " + FileSerialize.serialize2file(self.package)
        if self.opts is None:
            self.opts = ""
        self.opts += "--target " + self.module_site_package_path
        QPT_MEMORY.pip_tool.install_local_package(self.package,
                                                  version=self.version,
                                                  whl_dir=os.path.join(self.module_path, DOWN_PACKAGES_RELATIVE_PATH),
                                                  no_dependent=self.no_dependent,
                                                  opts=self.opts)


class OnlineInstallWhlOpt(SubModuleOpt):
    def __init__(self,
                 package: str,
                 version: str = None,
                 to_module_env_path=True,
                 to_python_env_version=DEFAULT_PACKAGE_FOR_PYTHON_VERSION,
                 no_dependent=False,
                 find_links: str = None,
                 opts: str = None):
        super().__init__(disposable=True)
        self.package = package
        self.to_module_env = to_module_env_path
        self.no_dependent = no_dependent
        self.find_links = find_links
        self.opts = opts
        self.version = version
        self.to_python_env_version = to_python_env_version
        if to_python_env_version:
            assert to_module_env_path, "安装在当前环境则不需要设置Python版本号参数to_python_env_version。" \
                                       "若需要安装其它位置，请设置to_module_env_path参数使包安装在其它位置。"
            self.to_python_env_version = DEFAULT_PACKAGE_FOR_PYTHON_VERSION

    def act(self) -> None:
        if "[FLAG-FileSerialize]" in self.package[:32]:
            self.package = self.package.strip("[FLAG-FileSerialize]")
            self.package = "-r " + FileSerialize.serialize2file(self.package)
        if self.to_module_env:
            if not self.opts:
                self.opts = ""
            self.opts += "--target " + self.module_site_package_path
            if self.to_python_env_version:
                self.opts += f" --python-version {self.to_python_env_version} --only-binary :all:"
        QPT_MEMORY.pip_tool.pip_package_shell(self.package,
                                              act="install",
                                              version=self.version,
                                              find_links=self.find_links,
                                              no_dependent=self.no_dependent,
                                              opts=self.opts)


class BatchInstallationOpt(SubModuleOpt):
    def __init__(self, path=None):
        super(BatchInstallationOpt, self).__init__(disposable=True)
        self.path = path

    def act(self) -> None:
        if self.path is None:
            self.path = os.path.join(self.module_path, DOWN_PACKAGES_RELATIVE_PATH)
        ready_list = PythonPackages.search_packages_dist_info()[0].keys()
        whl_list = [whl.split("-")[0] for whl in os.listdir(self.path)
                    if whl.split("-")[0] not in ready_list and
                    whl.split("-")[0].lower().replace("_", "-") not in ready_list]
        Logging.info(f"需要补充的安装包数量为：{len(whl_list)}")
        for whl_name in whl_list:
            QPT_MEMORY.pip_tool.install_local_package(whl_name,
                                                      whl_dir=self.path,
                                                      no_dependent=True)


class CustomPackage(SubModule):
    def __init__(self,
                 package,
                 version: str = None,
                 deploy_mode=None,
                 no_dependent=False,
                 find_links: str = None,
                 opts: str = None):
        super().__init__(name=None)
        if deploy_mode is None:
            deploy_mode = DEFAULT_DEPLOY_MODE
        if deploy_mode == LOCAL_DOWNLOAD_DEPLOY_MODE:
            self.add_pack_opt(DownloadWhlOpt(package=package,
                                             version=version,
                                             no_dependent=no_dependent,
                                             find_links=find_links,
                                             opts=opts))
            self.add_unpack_opt(LocalInstallWhlOpt(package=package,
                                                   version=version,
                                                   no_dependent=no_dependent,
                                                   opts=opts))
        elif deploy_mode == ONLINE_DEPLOY_MODE:
            self.add_unpack_opt(OnlineInstallWhlOpt(package=package,
                                                    version=version,
                                                    no_dependent=no_dependent,
                                                    find_links=find_links,
                                                    opts=opts))
        elif deploy_mode == LOCAL_INSTALL_DEPLOY_MODE:
            self.add_pack_opt(OnlineInstallWhlOpt(package=package,
                                                  version=version,
                                                  no_dependent=no_dependent,
                                                  find_links=find_links,
                                                  opts=opts))


class _RequirementsPackage(SubModule):
    def __init__(self,
                 requirements_file_path,
                 deploy_mode=None):
        super().__init__(name=None)
        if deploy_mode is None:
            deploy_mode = DEFAULT_DEPLOY_MODE

        fs_data = ""
        # 部分情况需要序列化requirement.txt文件
        if deploy_mode != LOCAL_INSTALL_DEPLOY_MODE:
            fs = FileSerialize(requirements_file_path)
            fs_data = "[FLAG-FileSerialize]" + fs.get_serialize_data()
        requirements_file_path = "-r " + requirements_file_path
        if deploy_mode == LOCAL_DOWNLOAD_DEPLOY_MODE:
            self.add_pack_opt(DownloadWhlOpt(package=requirements_file_path,
                                             no_dependent=False))
            self.add_unpack_opt(LocalInstallWhlOpt(package=fs_data,
                                                   no_dependent=False))
        elif deploy_mode == ONLINE_DEPLOY_MODE:
            self.add_unpack_opt(OnlineInstallWhlOpt(package=fs_data,
                                                    no_dependent=False))
        elif deploy_mode == LOCAL_INSTALL_DEPLOY_MODE:
            self.add_pack_opt(OnlineInstallWhlOpt(package=requirements_file_path,
                                                  no_dependent=False,
                                                  to_module_env_path=True))


class QPTDependencyPackage(SubModule):
    def __init__(self):
        self.level = TOP_LEVEL_REDUCE
        super().__init__(name=None)
        kernel_dependency_path = os.path.join(os.path.split(__file__)[0], "kernel_dependency.txt")
        lazy_dependency_path = os.path.join(os.path.split(__file__)[0], "qpt_lazy_dependency.txt")
        lazy_dependency_serialize = "[FLAG-FileSerialize]" + FileSerialize(lazy_dependency_path).get_serialize_data()
        kernel = "-r " + kernel_dependency_path
        lazy = "-r " + lazy_dependency_path
        self.add_pack_opt(OnlineInstallWhlOpt(package="qpt",
                                              version=qpt_version,
                                              no_dependent=True,
                                              to_module_env_path=True))
        self.add_pack_opt(OnlineInstallWhlOpt(package=kernel,
                                              no_dependent=False,
                                              to_module_env_path=True))
        self.add_pack_opt(DownloadWhlOpt(package=lazy,
                                         no_dependent=False))
        self.add_unpack_opt(LocalInstallWhlOpt(package=lazy_dependency_serialize,
                                               no_dependent=False))


class QPTGUIDependencyPackage(SubModule):
    def __init__(self):
        self.level = TOP_LEVEL_REDUCE
        super().__init__(name=None)
        kernel_dependency_path = os.path.join(os.path.split(__file__)[0], "kernel_dependency_GUI.txt")
        kernel = "-r " + kernel_dependency_path
        self.add_pack_opt(OnlineInstallWhlOpt(package=kernel,
                                              no_dependent=False,
                                              to_module_env_path=True))


class BatchInstallation(SubModule):
    def __init__(self):
        super().__init__()
        self.level = LOW_LEVEL
        if DEFAULT_DEPLOY_MODE == LOCAL_DOWNLOAD_DEPLOY_MODE:
            self.add_unpack_opt(BatchInstallationOpt())

# 自动推理依赖时需要特殊处理的Module配置列表 格式{包名: (Module, Module参数字典)}
# version、deploy_mode 为必填字段



import logging
import os

__all__ = ["Logging", "change_none_color", "clean_stout", "TProgressBar", "set_logger_file"]

formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')
logger = logging.getLogger("qpt_logger")
st_handler = logging.StreamHandler()
st_handler.setFormatter(formatter)

DEBUG_VAR = os.getenv("QPT_MODE")
if DEBUG_VAR == "Run":
    LEVEL = logging.INFO
elif DEBUG_VAR == "Debug":
    LEVEL = logging.DEBUG
else:
    LEVEL = logging.DEBUG

logger.setLevel(LEVEL)
st_handler.setLevel(LEVEL)
st_handler.terminator = ""

# 设置标准输出流
logger.addHandler(st_handler)


def set_logger_file(file_path):
    f_handler = logging.FileHandler(file_path, "w", encoding="utf-8")
    f_handler.setLevel(logging.DEBUG)
    f_handler.setFormatter(formatter)
    f_handler.terminator = ""
    logger.addHandler(f_handler)


def clean_stout(name_list: list):
    headers = logging.root.handlers
    for header in list(headers):
        if header.name in name_list:
            logging.root.removeHandler(header)


WARNING_SUMMARY = list()
ERROR_SUMMARY = list()


class BaseLogging:
    @staticmethod
    def final(clear=True):
        """
        用于打印当前收到的警告和报错情况
        :return: 是否包含报错或警告
        """
        Logging.info("-" * 10 + "WARNING SUMMARY")
        for msg in WARNING_SUMMARY:
            Logging.info(msg)
        Logging.info("-" * 10 + "ERROR SUMMARY  ")
        for msg in ERROR_SUMMARY:
            Logging.info(msg)
        Logging.info("-" * 10 + f"生成状态WARNING:{len(WARNING_SUMMARY)} ERROR:{len(ERROR_SUMMARY)}")
        flag = ERROR_SUMMARY.copy()
        if clear:
            WARNING_SUMMARY.clear()
            ERROR_SUMMARY.clear()
        if len(flag) > 0:
            return True
        else:
            return False

    @staticmethod
    def flush():
        st_handler.flush()


class LoggingColor(BaseLogging):
    @staticmethod
    def info(msg: str, line_feed=True):
        msg = "\033[34m" + msg + "\033[0m"
        if line_feed:
            msg += "\n"
        logger.info(msg)

    @staticmethod
    def debug(msg: str, line_feed=True):
        msg = "\033[35m" + msg + "\033[0m"
        if line_feed:
            msg += "\n"
        logger.debug(msg)

    @staticmethod
    def warning(msg: str, line_feed=True):
        msg = "\033[33m" + msg + "\033[0m"
        WARNING_SUMMARY.append(f"{len(WARNING_SUMMARY)}|{msg}")
        if line_feed:
            msg += "\n"
        logger.warning(msg)

    @staticmethod
    def error(msg: str, line_feed=True):
        msg = "\033[41m" + msg + "\033[0m"
        ERROR_SUMMARY.append(f"{len(WARNING_SUMMARY)}|{msg}")
        if line_feed:
            msg += "\n"
        logger.error(msg)


class LoggingNoneColor(BaseLogging):
    @staticmethod
    def info(msg: str, line_feed=True):
        if line_feed:
            msg += "\n"
        logger.info(msg)

    @staticmethod
    def debug(msg: str, line_feed=True):
        if line_feed:
            msg += "\n"
        logger.debug(msg)

    @staticmethod
    def warning(msg: str, line_feed=True):
        if line_feed:
            msg += "\n"
        logger.warning(msg)
        WARNING_SUMMARY.append(f"{len(WARNING_SUMMARY)}|{msg}\n")

    @staticmethod
    def error(msg: str, line_feed=True):
        if line_feed:
            msg += "\n"
        logger.error(msg)
        ERROR_SUMMARY.append(f"{len(WARNING_SUMMARY)}|{msg}\n")


def change_none_color():
    global Logging
    Logging = LoggingNoneColor


COLOR_VAR = os.getenv("QPT_COLOR")
if COLOR_VAR == "True" or COLOR_VAR is None:
    Logging = LoggingColor
else:
    change_none_color()


class TProgressBar:
    def __init__(self,
                 msg: str = "",
                 max_len: int = 100):
        self.count = 0
        self.msg = msg
        self.max_len = max_len - 1
        self.block = 20
        self.with_logging = True
        Logging.info("", line_feed=False)
        Logging.flush()

    def step(self, add_start_info="", add_end_info=""):
        self.count += 1
        if self.max_len == 0:
            block_str = "|" + "".join(['█'] * int(self.block)).ljust(self.block, " ") + "|"
            Logging.info("\r" +
                         self.msg +
                         f"\r1/1 {add_start_info} {block_str} {100:.2f}% " +
                         add_end_info,
                         line_feed=False)
        else:
            rate = self.count / self.max_len

            block_str = "|" + "".join(['█'] * int(rate * self.block)).ljust(self.block, " ") + "|"
            block = f"{block_str}"
            Logging.info("\r" +
                         self.msg +
                         f"\t{self.count}/{self.max_len} {add_start_info} {block} {rate * 100:.2f}% " +
                         add_end_info,
                         line_feed=False)
        Logging.flush()
        if self.count == self.max_len:
            Logging.info("", line_feed=True)
            Logging.flush()



# Author: Acer Zhang
# Datetime: 2021/5/26 
# Copyright belongs to the author.
# Please indicate the source for reprinting.
import os

from qpt.kernel.qos import dynamic_load_package, get_qpt_tmp_path
from qpt.kernel.qlog import clean_stout, Logging
from qpt.kernel.qterminal import PTerminal, TerminalCallback, LoggingTerminalCallback
from qpt.kernel.qcode import PythonPackages

TSINGHUA_PIP_SOURCE = "https://pypi.tuna.tsinghua.edu.cn/simple"
BAIDU_PIP_SOURCE = "https://mirror.baidu.com/pypi/simple"
DOUBAN_PIP_SOURCE = "https://pypi.douban.com/simple"
BFSU_PIP_SOURCE = "https://mirrors.bfsu.edu.cn/pypi/web/simple"
DEFAULT_PIP_SOURCE = BFSU_PIP_SOURCE


class PIPTerminal(PTerminal):
    # ToDo 此处需要简化
    def __init__(self, python_path):
        super(PIPTerminal, self).__init__()
        self.head = os.path.abspath(python_path) + " -m pip"

    def shell_func(self, callback: TerminalCallback = LoggingTerminalCallback()):
        def closure(closure_shell):
            if isinstance(closure_shell, list):
                tmp = [" " + bit_shell for bit_shell in closure_shell]
                closure_shell = "".join(tmp)
            closure_shell = self.head + closure_shell
            if closure_shell[1:3] == ":\\":
                closure_shell = f"cd {closure_shell[:2]} ;" + closure_shell
            Logging.debug(f"SHELL: {closure_shell}")
            closure_shell = f'{closure_shell}; echo "---QPT OUTPUT STATUS CODE---" ;$? \n'
            # 发送指令
            try:
                final_shell = closure_shell.encode("utf-8")
            except Exception as e:
                Logging.error("执行该指令时遇到解码问题，目前将采用兼容模式进行，原始报错如下：\n" + str(e))
                final_shell = closure_shell.encode("utf-8", errors="ignore")
            self.main_terminal.stdin.write(final_shell)
            try:
                self.main_terminal.stdin.flush()
            except Exception as e:
                Logging.error(str(e))

            # 捕获PIP的异常
            callback.error_fitter = ["ERROR:"]
            callback.normal_fitter = ["requires click, which is not installed.\nqpt",
                                      "ERROR: pip's dependency resolver "]
            callback.handle(self.main_terminal)

        return closure


class PipTools:
    """
    Python解释器管理器
    """

    def __init__(self,
                 source: str = None,
                 pip_path=None):
        if source is None:
            source = DEFAULT_PIP_SOURCE
        if pip_path:
            pip_main = dynamic_load_package(packages_name="pip", lib_packages_path=pip_path).main
        else:
            from pip._internal.cli.main import main as pip_main
        self.pip_main = pip_main
        self.pip_local = pip_main
        self.source = source

        # 安静模式
        self.quiet = True if os.getenv("QPT_MODE") == "Run" else False
        if self.quiet:
            Logging.debug("PipTools进入安静模式")

        # ToDo 可考虑增加环境管理部分 - 可考虑生成软链
        pass

    def pip_shell(self, shell):
        if not os.path.exists(get_qpt_tmp_path('pip_cache')):
            os.makedirs(get_qpt_tmp_path('pip_cache'), exist_ok=True)
        shell += f" --isolated --disable-pip-version-check --cache-dir {get_qpt_tmp_path('pip_cache')}" \
                 f" --timeout 10"
        if self.quiet:
            shell += " --quiet"
        self.pip_main(shell.split(" "))
        clean_stout(['console', 'console_errors', 'console_subprocess'])

    def pip_package_shell(self,
                          package: str = None,
                          version: str = None,
                          act="install",
                          no_dependent=False,
                          find_links: str = None,
                          opts: str = None):
        if version:
            package += "==" + version

        shell = f"{act} {package} -i {self.source}"

        if no_dependent:
            shell += " --no-deps"

        if find_links:
            shell += " -f " + find_links

        if opts:
            shell += " " + opts

        self.pip_shell(shell)

    def download_package(self,
                         package: str,
                         save_path: str,
                         version: str = None,
                         no_dependent=False,
                         find_links: str = None,
                         python_version: str = None,
                         opts: str = None):
        d_opts = "-d " + save_path
        if python_version:
            d_opts += " --python-version " + python_version
            d_opts += " --only-binary :all:"

        if opts:
            d_opts += " " + opts

        # pip download xxx -d ./test
        # pip install xxx -f ./test --no-deps
        self.pip_package_shell(package=package,
                               version=version,
                               act="download",
                               no_dependent=no_dependent,
                               find_links=find_links,
                               opts=d_opts)

    def install_local_package(self,
                              package: str,
                              version: str = None,
                              whl_dir: str = None,
                              no_dependent=False,
                              opts: str = None):
        i_opts = "--no-index"
        if opts:
            i_opts += " " + opts
        self.pip_package_shell(package=package,
                               act="install",
                               version=version,
                               no_dependent=no_dependent,
                               find_links=whl_dir,
                               opts=i_opts)

    def analyze_dependence(self, analyze_path, save_file_path=None, return_path=False):
        if save_file_path is None:
            save_file_path = os.path.join(analyze_path, "requirements_with_opt.txt")

        requires, dep, ignore_requires = PythonPackages.intelligent_analysis(analyze_path, return_all_info=True)

        with open(save_file_path, "w", encoding="utf-8") as req_file:
            req_file.write("# Here is the list of packages automatically derived by QPT\n"
                           "# you can ignore the dependent packages in the main package and only care "
                           "about the main package\n"
                           "# For example, you need to install paddlepaddle and pillow, because paddlepaddle "
                           "relies on pillow, so you only need to install paddlepaddle.\n"
                           "# ---------------------------------------------------------------------\n"
                           "# QPT Home:        https://github.com/GT-ZhangAcer/QPT\n"
                           "# ---------------------------------------------------------------------\n"
                           "# \n"
                           "# -------------Mainly depends on package analysis results--------------\n\n")
            temp_write = "\n# ----------------------Ignored sub dependent packages---------------------\n"
            for require_name, require_version in requires.items():
                req_file.write(f"{require_name}=={require_version}\n")
                if dep[require_name]:
                    temp_write += f"\n# -----Dependencies of {require_name}\n"
                    for dep_name, dep_version in dep[require_name].items():
                        if not dep_version:
                            dep_version = ""
                        temp_write += f"#{dep_name}{dep_version}\n"

            req_file.write("\n# ----------------------Ignored dependent packages---------------------\n")
            for ignore_require_name, ignore_require_version in ignore_requires.items():
                req_file.write(f"#{ignore_require_name}=={ignore_require_version}\n")
            req_file.write(temp_write)

        Logging.info(f"依赖分析完毕!\n"
                     f"已在\033[32m{os.path.abspath(save_file_path)}\033[0m 中创建了依赖列表\n"
                     f"Tips 1: 查看文件后可能需要关闭查看该文件的文本查看器，这样可以有效避免文件被占用\n"
                     f"Tips 2: 请务必检查上方文件中所写入的依赖列表情况，因为自动分析并不能保证程序依赖均可以被检出\n"
                     f"        若在执行EXE时提示:ImportError: No module named xxx 报错信息，请在该依赖文件中加入xxx或取消xxx前的 # 符号\n"
                     f"---------------------------------------------------------------------\n"
                     f"\033[41m请在检查/修改依赖文件后\033[0m在此处按下回车键继续...\n"
                     f"请键入指令[回车键 - 一次不行可以试试按两次]:_", line_feed=False)
        input()
        if return_path:
            return save_file_path
        else:
            return self.analyze_requirements_file(save_file_path)

    @staticmethod
    def analyze_requirements_file(file_path):
        requirements = dict()
        try:
            with open(file_path, "r", encoding="utf-8") as req_file:
                data = req_file.readlines()
                for line in data:
                    line = line.strip("\n")
                    if "==" in line:
                        package, version = line.split("==")
                    else:
                        package = line
                        version = None
                    requirements[package] = version
        except Exception as e:
            raise Exception(f"{file_path}文件解析失败，文件可能被其他程序占用或格式异常\n"
                            f"报错信息如下：{e}")
        return requirements

    @staticmethod
    def save_requirements_file(requirements_dict, save_file_path, encoding="utf-8"):
        with open(save_file_path, "w", encoding=encoding) as file:
            for requirement in requirements_dict:
                if requirements_dict[requirement] is not None:
                    line = f"{requirement}=={requirements_dict[requirement]}\n"
                else:
                    line = requirement + "\n"
                file.write(line)



# Author: Acer Zhang
# Datetime: 2021/5/26
# Copyright belongs to the author.
# Please indicate the source for reprinting.

import os
import sys
import shutil
import base64
import datetime
import tempfile

from typing import List

import qpt
from qpt.memory import QPT_MEMORY
from qpt.version import version as qpt_v
from qpt.modules.base import SubModule
from qpt.modules.python_env import BasePythonEnv, AutoPythonEnv
from qpt.modules.package import QPTDependencyPackage, QPTGUIDependencyPackage, \
    DEFAULT_DEPLOY_MODE, \
    set_default_deploy_mode, BatchInstallation
from qpt.modules.auto_requirements import AutoRequirementsPackage

from qpt.kernel.qlog import Logging, TProgressBar, set_logger_file
from qpt.kernel.qos import clean_qpt_cache, copytree, check_chinese_char, StdOutLoggerWrapper, warning_msg_box
from qpt.kernel.qterminal import PTerminal, RunTerminalCallback
from qpt.smart_opt import set_default_pip_lib
from qpt.memory import QPT_MODE, check_all, get_env_vars, CheckRun
from qpt.kernel.qpe import make_icon
from qpt.gui.tk_progressbar import ProgressbarFrame

__all__ = ["CreateExecutableModule", "RunExecutableModule"]


class CreateExecutableModule:
    def __init__(self,
                 work_dir,
                 launcher_py_path,
                 save_path,
                 ignore_dirs: list = None,
                 requirements_file="auto",
                 icon=None,
                 deploy_mode=DEFAULT_DEPLOY_MODE,
                 sub_modules: List[SubModule] = None,
                 interpreter_module: BasePythonEnv = None,
                 hidden_terminal: bool = False,
                 with_debug: bool = False):
        self.with_debug = with_debug

        # 初始化路径成员变量
        self.launcher_py_path = os.path.relpath(launcher_py_path, work_dir)
        if self.launcher_py_path[:1] == "\\":
            self.launcher_py_path = self.launcher_py_path[1:]
        #         assert "." not in self.launcher_py_path.strip(".py"), \
        #             f"{self.launcher_py_path}中的路径或文件名中出现了除“.py”以外的“.”符号，请保证路径和文件中没有除“.py”以外的“.”符号\n" \
        #             f"例如：C:/123.445/run.py 中123.445文件夹包含了“.”符号，该符号将可能导致Python程序运行终止，请修改该类情况！"
        assert " " not in self.launcher_py_path, \
            f"{self.launcher_py_path}中的路径或文件名中出现了空格符号，请删去文件夹或文件名中的空格\n" \
            f"例如：C:/123 445/run.py 中123 445文件夹包含了空格符号，该符号将可能导致Python程序运行终止，请修改该类情况！"

        self.work_dir = work_dir
        assert os.path.exists(os.path.join(self.work_dir, self.launcher_py_path)), \
            f"请检查{launcher_py_path}文件是否存在{self.work_dir}目录"

        self.save_path = save_path
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path, exist_ok=True)
        self.module_path = os.path.join(save_path, "Release")
        self.debug_path = os.path.join(save_path, "Debug")
        self.interpreter_path = os.path.join(self.module_path, "Python")
        if ignore_dirs is None:
            self.ignore_dirs = list()

        self.icon_path = icon

        # 创建基本环境目录
        if os.path.exists(self.save_path):
            if os.path.exists(self.module_path):
                Logging.warning(f"{os.path.abspath(self.module_path)}已存在，已清空该目录")
                self.ignore_dirs.append(self.module_path)
                shutil.rmtree(self.module_path)
            if os.path.exists(self.debug_path):
                Logging.warning(f"{os.path.abspath(self.debug_path)}已存在，已清空该目录")
                self.ignore_dirs.append(self.debug_path)
                shutil.rmtree(self.debug_path)
        else:
            os.makedirs(self.save_path, exist_ok=True)

        # 配置操作参数
        set_default_deploy_mode(deploy_mode)
        set_default_pip_lib(self.interpreter_path)
        self.with_debug = with_debug
        self.hidden_terminal = hidden_terminal

        # 新建配置信息
        self.configs = dict()
        self.configs["launcher_py_path"] = self.launcher_py_path
        self.configs["hidden_terminal"] = hidden_terminal
        self.configs["lazy_module"] = list()
        self.configs["sub_module"] = list()
        self.configs["local_uid"] = base64.b64encode((os.path.abspath(sys.executable) + "|" +
                                                      os.path.abspath(qpt.__file__) + "|" +
                                                      os.path.abspath(self.module_path)).encode('utf-8'))

        # 额外的成员变量
        self.resources_path = os.path.join(self.module_path, "resources")
        self.config_path = os.path.join(self.module_path, "configs")
        self.config_file_path = os.path.join(self.config_path, "configs.txt")
        self.lib_package_path = os.path.join(self.interpreter_path,
                                             QPT_MEMORY.site_packages_path)

        # 设置全局下载的Python包默认解释器版本号 - 更换兼容性方案
        # set_default_package_for_python_version(interpreter_module.python_version)

        # 避免打包虚拟环境等
        venv_dir = "-%NONE-FLAG%-"
        for root, dirs, files in os.walk(self.work_dir):
            if "pyvenv.cfg" in files:
                venv_dir = root
                self.ignore_dirs.append(root)
                Logging.warning(f"检测到pyvenv.cfg，推测出{os.path.abspath(root)}为Python虚拟环境主目录，在打包时会忽略该目录")
                continue
            if ".github" in dirs:
                self.ignore_dirs.append(os.path.join(root, ".github"))
                Logging.warning(f"检测到.github，推测出{os.path.abspath(root)}为.github目录，在打包时会忽略该目录")
            if ".git" in dirs:
                self.ignore_dirs.append(os.path.join(root, ".git"))
                Logging.warning(f"检测到.git，推测出{os.path.abspath(root)}为.git目录，在打包时会忽略该目录")
            if ".idea" in dirs:
                self.ignore_dirs.append(os.path.join(root, ".idea"))
                Logging.warning(f"检测到.idea，推测出{os.path.abspath(root)}为.idea目录，在打包时会忽略该目录")
            if "__pycache__" in dirs and not (venv_dir in root and root.index(venv_dir) == 0):
                self.ignore_dirs.append(os.path.join(root, "__pycache__"))

        # Module相关
        # 初始化解释器Module
        if interpreter_module is None:
            interpreter_module = AutoPythonEnv()
        # 获取SubModule列表 - 此处均无ExtModule
        self.lazy_modules = [interpreter_module, QPTDependencyPackage()]

        self.sub_modules = sub_modules if sub_modules is not None else list()

        # 放入增强包
        self.add_sub_module(BatchInstallation())
        # 放入增强包
        if self.hidden_terminal:
            self.add_sub_module(QPTGUIDependencyPackage())

        # 解析依赖
        if requirements_file == "auto":
            auto_dependency_module = AutoRequirementsPackage(path=self.work_dir,
                                                             deploy_mode=deploy_mode)
        else:
            auto_dependency_module = AutoRequirementsPackage(path=requirements_file,
                                                             deploy_mode=deploy_mode)
        self.add_sub_module(auto_dependency_module)

        # 初始化终端 - 占位 待lazy_module执行完毕后生成终端（依赖Qt lazy module）
        self.terminal = None

    def add_sub_module(self, sub_module: SubModule, lazy=False):
        """
        为Module添加子模块
        """
        module = sub_module.get_all_module()
        if len(module) > 1:
            for em in module[1:]:
                self.add_sub_module(em)
        if lazy:
            self.lazy_modules += [sub_module]
        else:
            self.sub_modules.append(sub_module)

    def print_details(self):
        Logging.info("----------QPT执行使用了以下OP----------")
        for module in self.lazy_modules:
            Logging.info(module.__class__.__name__ + f"执行优先级{module.level}" + f"\t{module.details}")
        Logging.info("----------程序执行使用了以下OP----------")
        for module in self.sub_modules:
            Logging.info(module.__class__.__name__ + f"优先级{module.level}" + f"\t{module.details}")
        Logging.info("------------------------------------")

    def _solve_module(self, lazy=False):
        if lazy:
            modules = self.lazy_modules
            # lazy mode 不支持terminal
            terminal = None
        else:
            self.terminal = PTerminal()
            modules = self.sub_modules
            terminal = self.terminal.shell_func()
        for sub in modules:
            # ToDO设置序列化路径
            sub._module_path = self.module_path
            # 需对每个module设置save_dir和终端
            sub.prepare(work_dir=self.work_dir,
                        interpreter_path=os.path.join(self.module_path, "Python"),
                        module_path=self.module_path,
                        terminal=terminal)
            sub.pack()

            # 保护用户侧接触不到的模块不被泄漏模块名
            if len(sub.unpack_opts) != 0:
                if lazy:
                    self.configs["lazy_module"].append(sub.name)
                else:
                    self.configs["sub_module"].append(sub.name)

    def make(self):
        # 打印sub module信息
        # 依靠优先级进行排序
        self.lazy_modules.sort(key=lambda m: m.level, reverse=True)
        self.sub_modules.sort(key=lambda m: m.level, reverse=True)
        self.print_details()

        # 解析子模块
        self._solve_module(lazy=True)
        self._solve_module()

        # 复制资源文件
        assert os.path.exists(self.work_dir), f"{os.path.abspath(self.work_dir)}不存在，请检查该路径是否正确"
        Logging.info("正在复制相关文件，可能会耗时较长")
        copytree(self.work_dir, self.resources_path, ignore_dirs=self.ignore_dirs)

        # QPT的dev模式
        if self.with_debug:
            Logging.debug("当前已开启QPT-dev模式，将会复制当前版本的QPT文件至相应目录")
            qpt_dir_path = os.path.split(qpt.__file__)[0]
            copytree(src=qpt_dir_path, dst=os.path.join(self.interpreter_path, "Lib/site-packages/qpt"))

        # # 避免出现if __name__ == '__main__':
        # with open(os.path.join(self.resources_path, self.launcher_py_path), "r", encoding="utf-8") as lf:
        #     lf_codes = lf.readlines()
        # for lf_code_id, lf_code in enumerate(lf_codes):
        #     if "if" in lf_code and "__name__" in lf_code and "__main__" in lf_code:
        #         # 懒得写正则了嘿嘿嘿
        #         Logging.warning(f"{self.launcher_py_path}中包含if __name__ == '__main__'语句，"
        #                         f"由于用户使用时QPT成为了主程序，故此处代码块会被Python忽略。"
        #                         f"为保证可以正常执行，当前已自动修复该问题")
        #         with open(os.path.join(self.resources_path, self.launcher_py_path), "w", encoding="utf-8") as new_lf:
        #             lf_codes[lf_code_id] = lf_code[:lf_code.index("if ")] + "if 'qpt':\n"
        #             new_lf.writelines(lf_codes)

        # 创建配置文件
        os.makedirs(self.config_path, exist_ok=True)
        with open(self.config_file_path, "w", encoding="utf-8") as config_file:
            config_file.write(str(self.configs))

        # 启动器相关
        launcher_entry_path = os.path.join(os.path.split(qpt.__file__)[0], "ext/launcher_entry")
        # 复制Debug所需文件
        Logging.info("正在复制相关文件，可能会耗时较长")
        debug_ext_dir = os.path.join(os.path.split(qpt.__file__)[0], "ext/launcher_debug")
        copytree(debug_ext_dir, dst=self.debug_path)
        copytree(self.module_path, dst=self.debug_path)
        shutil.copy(src=os.path.join(launcher_entry_path, "entry_debug.cmd"),
                    dst=os.path.join(self.debug_path, "configs/entry.cmd"))
        # 生成Debug标识符
        unlock_file_path = os.path.join(self.debug_path, "configs/unlock.cache")
        with open(unlock_file_path, "w", encoding="utf-8") as unlock_file:
            unlock_file.write(str(datetime.datetime.now()))
        # 重命名兼容模式文件
        compatibility_mode_file = os.path.join(self.debug_path, "compatibility_mode.cmd")
        if os.path.exists(compatibility_mode_file):
            os.rename(compatibility_mode_file,
                      os.path.join(self.debug_path, "使用兼容模式运行.cmd"))

        # 复制Release启动器文件
        launcher_ext_dir = os.path.join(os.path.split(qpt.__file__)[0], "ext/launcher")
        launcher_ignore_file = None
        if self.hidden_terminal:
            shutil.copy(src=os.path.join(launcher_entry_path, "entry_run_hidden.cmd"),
                        dst=os.path.join(self.module_path, "configs/entry.cmd"))
        else:
            launcher_ignore_file = ["Main.exe"]
            shutil.copy(src=os.path.join(debug_ext_dir, "Debug.exe"), dst=os.path.join(self.module_path, "启动程序.exe"))
            shutil.copy(src=os.path.join(launcher_entry_path, "entry_run.cmd"),
                        dst=os.path.join(self.module_path, "configs/entry.cmd"))

        copytree(launcher_ext_dir, dst=self.module_path, ignore_files=launcher_ignore_file)
        # 重命名兼容模式文件
        compatibility_mode_file = os.path.join(self.module_path, "compatibility_mode.cmd")
        if os.path.exists(compatibility_mode_file):
            os.rename(compatibility_mode_file,
                      os.path.join(self.module_path, "使用兼容模式运行.cmd"))
        # 重命名启动器文件
        launcher_file = os.path.join(self.module_path, "Main.exe")
        if os.path.exists(launcher_file):
            os.rename(launcher_file,
                      os.path.join(self.module_path, "启动程序.exe"))
        # 修改icon
        if self.icon_path:
            make_icon(self.icon_path,
                      pe_path=os.path.join(self.module_path, "启动程序.exe"),
                      img_save_path=os.path.join(self.config_path, "Logo.ico"))
            make_icon(self.icon_path,
                      pe_path=os.path.join(self.debug_path, "Debug.exe"),
                      img_save_path=os.path.join(self.debug_path, "configs", "Logo.ico"))

        # Logging Summary
        if Logging.final():
            Logging.warning("SUMMARY结束，发现上述异常情况，请确认后按任意键继续！")
            Logging.flush()
            input()

        # 收尾工作
        Logging.info(f"\n制作完毕，保存位置为：{os.path.abspath(self.module_path)}，该目录下将会有以下文件夹\n"
                     f"| ----------------------------------------------------------------------------- |\n"
                     f"| Debug目录：\t该目录下提供了Debug环境，可简单验证打包后程序是否可以正常执行。        \n"
                     f"| Release目录：\t将该目录进行压缩，并发给您的用户，待您的用户打开该压缩包下的“启动程序.exe”后\n"
                     f"|             \t即可启动您制作的程序  \n"
                     f"| ----------------------------------------------------------------------------- |\n")
        Logging.warning(f"\n| ---------------------------------Warning!------------------------------------ |\n"
                        f"| 请勿在本机打开Release目录下的“启动程序.exe”文件，原因如下： \n"
                        f"| 1. 该程序会加载“一次性部署模块”，部署后该模块会消失，消失后可能无法在其他电脑上使用。\n"
                        f"| 2. 该程序会解压缩当前环境，执行“启动程序.exe”后整个目录大小可能会增加1~5倍。（取决于压缩率）\n"
                        f"| 3. 若需要测试打包后程序是否可以正常运行，请在Debug目录下进行测试。\n"
                        f"| 4. 若特殊情况必须在Release目录下进行测试，请制作Release目录的备份，在他人需要时提供该备份\n"
                        f"|    文件或重新打包，以避免因执行“启动程序.exe”后丢失“一次性部署模块”，从而无法被他人使用。\n"
                        f"| ----------------------------------------------------------------------------- |\n")

        sys.stdout.flush()
        Logging.info("是否需要保留QPT在打包时产生的缓存文件？若不清空则可能会在下次使用QPT时复用缓存以提升打包速度")
        Logging.info("[保留(Y)/清空(N)]:_", line_feed=False)
        clear_key = input()
        sys.stdout.flush()
        if clear_key.lower() == "n":
            clean_qpt_cache()
            Logging.info("QPT缓存已全部清空")
        os.startfile(os.path.abspath(self.save_path))


class RunExecutableModule:
    def __init__(self, module_path):
        # 初始化Module信息
        self.base_dir = os.path.abspath(module_path)
        self.config_path = os.path.join(self.base_dir, "configs")
        self.config_file_path = os.path.join(self.base_dir, "configs", "configs.txt")
        self.work_dir = os.path.join(self.base_dir, "resources")
        self.interpreter_path = os.path.join(self.base_dir, "Python")

        # 初始化Log
        log_name = str(datetime.datetime.now()).replace(" ", "_").replace(":", "-") + ".txt"
        if not os.path.exists(os.path.join(self.config_path, "logs")):
            log_name = "First-" + log_name
            os.mkdir(os.path.join(self.config_path, "logs"))
        if QPT_MODE == "Debug":
            log_name = "#Debug#" + log_name
        set_logger_file(os.path.join(self.config_path, "logs", "QPT-" + log_name))
        sys.stdout = StdOutLoggerWrapper(os.path.join(self.config_path, "logs", "APP-" + log_name))

        # 系统信息
        check_all()

        # 软件信息
        Logging.info(f"QPT Runtime版本号为{qpt_v}，若无法使用该程序，可向程序发布者或GitHub: QPT-Family/QPT提交issue寻求帮助")

        # 强制本地PIP
        set_default_pip_lib(self.interpreter_path)

        # 检查路径是否非法
        check_path = __file__
        if os.path.realpath(tempfile.gettempdir()) in os.path.realpath(self.base_dir):
            warning_msg_box(text=f"{self.base_dir}\n"
                                 f"上述目录存在于系统的临时目录下，该情况可能会对程序运行造成影响\n"
                                 f"建议的解决方案如下：\n"
                                 f"1. 请勿在压缩软件中打开本程序，务必解压后再运行。\n"
                                 f"2. 请在物理硬盘上执行本程序。")
        if check_chinese_char(check_path) or " " in check_path:
            warning_msg_box(text=f"{self.base_dir}\n"
                                 f"警告！当前路径↑中包含中文或空格，部分软件包将无法运行，强烈建议您修改相关的文件夹名，\n"
                                 f"---------------------------------------\n"
                                 f"不符合规范的路径如下：\n"
                                 f"C:/GT真菜/xxx/yyy\r\r！“真菜”为中文\n"
                                 f"C:/zzz/GT真 菜/yyy\r\r！“GT真 菜”中“真”字后带有空格\n"
                                 f"---------------------------------------\n"
                                 f"符合规范的路径如下：\n"
                                 f"C:/hello/xxx/yyy\n"
                                 f"---------------------------------------\n"
                                 f"当然，您也可将Windows系统的默认编码模式更改为UTF-8，这可以更好兼容中文，但操作难度较高。\n"
                                 f"---------------------------------------\n"
                                 f"请修改相关文件名后重新运行，谢谢！",
                            force=True)
            Logging.info("程序已停止")
            exit(1)

        try:
            with open(self.config_file_path, "r", encoding="utf-8") as config_file:
                self.configs = eval(config_file.read())
        except Exception as e:
            Logging.error("请检查杀毒软件、防火墙等限制策略，当前程序无法正常访问Config.gt文件，完整报错如下：\n" + str(e))

        # 获取GUI选项
        self.hidden_terminal = self.configs["hidden_terminal"]

        # 获取Module
        self.lazy_module = self.configs["lazy_module"]
        self.sub_module = self.configs["sub_module"]

        # 实例化终端
        self.auto_terminal = PTerminal()

    def _solve_module(self):
        modules = self.lazy_module + self.sub_module

        def render(arg=None):
            Logging.info("初次使用将会适应本地环境，可能需要几分钟时间，请耐心等待...")
            terminal = self.auto_terminal.shell_func()
            tp = TProgressBar("初始化进度", max_len=len(modules) + 2)
            for sub_module_id, sub_name in enumerate(modules):
                tp.step(add_end_info=f"{sub_name}部署中...")
                if arg:
                    arg.step(text="正在适配" + sub_name)
                sub_module = SubModule(sub_name)
                sub_module.prepare(work_dir=self.work_dir,
                                   interpreter_path=self.interpreter_path,
                                   module_path=self.base_dir,
                                   terminal=terminal)
                sub_module.unpack()
            tp.step(add_end_info=f"初始化完毕")
            if arg:
                arg.close()

        if self.hidden_terminal:
            ProgressbarFrame(bind_fuc=render, max_step=len(modules))
        else:
            render()

    def solve_work_dir(self):
        # ToDo 加个Lock 彻底去除非Python的环境变量
        # Set Sys ENV
        sys.path.append(self.work_dir)
        sys.path.append(os.path.abspath("./Python/Lib/site-packages"))
        sys.path.append(os.path.abspath("./Python/Lib/ext"))
        sys.path.append(os.path.abspath("./Python/Lib"))
        sys.path.append(os.path.abspath("./Python"))
        sys.path.append(os.path.abspath("./Python/Scripts"))

        # Set PATH ENV
        env_vars = get_env_vars(self.work_dir)
        for k, v in env_vars.items():
            os.environ[k] = v

        # change dir
        os.chdir(self.work_dir)

    def run(self):
        # 设置工作目录
        self.solve_work_dir()

        # 获取启动信息 - 避免在Release下进行Debug
        env_warning_flag = False
        local_uid = base64.b64decode(self.configs["local_uid"]).decode("utf-8")
        if all([os.path.exists(uid) for uid in local_uid.split("|")]):
            env_warning_flag = True
        lock_file_path = os.path.join(self.config_path, "unlock.cache")
        if os.path.exists(lock_file_path):
            env_warning_flag = False

        if env_warning_flag:
            msg = warning_msg_box("Warning", f"非常不建议在该环境下进行调试，原因如下： \n"
                                             f" 1. 继续执行将会加载“一次性部署模块”，部署后该模块会消失，消失后可能无法在其他电脑上使用。\n"
                                             f" 2. 该程序会解压缩当前环境，执行“启动程序.exe”后整个目录大小可能会增加1~5倍。（取决于压缩率）\n"
                                             f" 3. 若需要测试打包后程序是否可以正常运行，请在Debug目录下进行测试。\n"
                                             f" 4. 若特殊情况必须在Release目录下进行测试，请制作Release目录的备份，在他人需要时提供该备份\n"
                                             f"    文件或重新打包，以避免因执行“启动程序.exe”后丢失“一次性部署模块”，从而无法被他人使用。\n"
                                             f"---------------------------------------------------------------------------\n"
                                             f"当然，这个警告框并不会在其它电脑上弹出，仅会在本计算机运行时提示"
                                             f"请问是否还要在该环境下继续执行？")
            if not msg:
                Logging.info("程序已停止")
                exit(1)
        # prepare module - GUI组件需要在此之后才能进行
        self._solve_module()

        if Logging.final():
            msg = warning_msg_box(title="ERROR 发生异常 -GitHub: QPT-Family/QPT",
                                  text="检测到安装中出现问题，若您为未成功运行过本程序，请点击 取消 \n"
                                       "若您已经成功运行了本程序，可点击 确定 来隐藏该窗口")
            if msg:
                CheckRun.make_run_file(self.config_path)
        else:
            CheckRun.make_run_file(self.config_path)

        # 执行主程序
        run_shell = f'cd "{self.work_dir}"' + \
                    '; ' + './../Python/python.exe "' + \
                    os.path.abspath(self.configs["launcher_py_path"]) + '"'
        # start -NoNewWindow 看看要不要从输入符号入手
        self.auto_terminal.shell(run_shell, callback=RunTerminalCallback())
        # main_lib_path = self.configs["launcher_py_path"].replace(".py", "")
        # main_lib_path = main_lib_path. \
        #     replace(".py", ""). \
        #     replace(r"\\", "."). \
        #     replace("\\", "."). \
        #     replace("/", ".")
        # # ToDo 等日志系统做好了再取消注释
        # os.system('cls')
        # lib = importlib.import_module(main_lib_path)
        # input("QPT执行完毕，请按任意键退出")



# Author: Acer Zhang
# Datetime:2021/8/29 
# Copyright belongs to the author.
# Please indicate the source for reprinting.

import os

from qpt.kernel.qos import get_qpt_tmp_path


def make_icon(ico_path, pe_path, save_path=None, img_save_path=None):
    if img_save_path:
        img_save_path = os.path.join(get_qpt_tmp_path("icon"), "make_icon.ico")

    if save_path is None:
        save_path = pe_path

    # Todo 后期加warmup
    import pefile
    import PIL.Image as Image
    # 处理图像
    img = Image.open(ico_path).resize((128, 128))
    img.save(img_save_path, sizes=[(128, 128)])

    # 读取图像
    assert os.path.exists(img_save_path), f"{os.path.abspath(img_save_path)}图像文件不存在，请检查路径拼写是否正确！"
    with open(img_save_path, "rb") as f:
        img_b = f.read()

    # PE
    pe = pefile.PE(pe_path)
    # offset 可从ExeScope中获取
    rva = pe.get_rva_from_offset(0x1D6E8)
    pe.set_bytes_at_rva(rva, img_b[22:])
    f = pe.write()
    with open(save_path, 'wb') as final_f:
        final_f.write(f)


if __name__ == '__main__':
    test1 = r'M:\ICON\i.ico'
    pe_test = r"M:\ICON\o2.exe"
    save_1 = r"M:\ICON\new1.exe"
    make_icon(test1, pe_test, save_path=save_1)



import shutil
import os
import sys
import tempfile
import io
from importlib import util

from qpt.kernel.qlog import Logging, TProgressBar
from qpt.memory import QPT_MEMORY
from qpt.version import version


def check_chinese_char(text):
    """
    判断是否包含中文，避免来自部分内鬼C++底层的Python包无缘无故报错
    :param text: 字符串
    :return: 是否包含
    """
    return not all(ord(char) < 128 for char in text)


TMP_BASE_PATH = os.path.join(tempfile.gettempdir(), f"QPT_Cache_V/{version}")
# Check User name is chinese
if check_chinese_char(TMP_BASE_PATH) or " " in TMP_BASE_PATH:
    TMP_BASE_PATH = "C:/q_tmp"
    Logging.warning(f"当前系统的用户名中包含中文/空格等可能会对程序造成异常的字符，现已默认QPT临时目录为{TMP_BASE_PATH}")
    os.makedirs(TMP_BASE_PATH, exist_ok=True)


def dynamic_load_package(packages_name, lib_packages_path):
    """
    动态加载Python包
    :param packages_name: 包名
    :param lib_packages_path: site-packages路径/包所在的目录
    :return: Python包
    """
    module_spec = util.find_spec(packages_name, lib_packages_path)
    module = util.module_from_spec(module_spec)
    module_spec.loader.exec_module(module)
    return module


def add_ua():
    """
    获取UA权限
    """
    import ctypes
    ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, __file__, None, 1)
    Logging.info("UA请求完毕")


def set_qpt_env_var(path):
    # ToDO:可考虑用Win32代替
    out = os.system(f'setx "QPT_BASE" {path} /m')
    if out == 0:
        return True
    else:
        return False


def download(url, file_name, path=None, clean=False):
    """
    下载指定文件至目录
    :param url: 下载的URL
    :param file_name: 保存的文件名
    :param path: 保存路径
    :param clean: 是否情况旧的下载数据
    :return: 1代表全新数据，0代表使用缓存
    """
    import wget
    if not os.path.exists(path):
        os.makedirs(path)
    file_path = os.path.join(path, file_name)
    if not os.path.exists(file_path) or clean:
        try:
            wget.download(url, file_path)
            return 1
        except Exception as e:
            Logging.error(f"无法下载文件，请检查网络是否可以连接以下链接\n"
                          f"{url}\n"
                          f"若该文件由QPT提供，请升级QPT版本，若版本升级后仍未解决可在以下地址提交issue反馈该情况\n"
                          f"https://github.com/GT-ZhangAcer/QPT/issues")
            raise Exception("文件下载失败，报错如下：" + str(e))
    else:
        return 0


def get_qpt_tmp_path(dir_name="Cache", clean=False):
    """
    获取一个临时目录
    :param dir_name: 临时目录名
    :param clean: 是否强制清空目录
    :return: 目录路径
    """
    dir_path = os.path.join(TMP_BASE_PATH, dir_name)
    if os.path.exists(dir_path) and clean:
        shutil.rmtree(dir_path)
        os.makedirs(dir_path, exist_ok=True)
    else:
        os.makedirs(dir_path, exist_ok=True)
    return dir_path


def clean_qpt_cache():
    dir_path = TMP_BASE_PATH
    shutil.rmtree(dir_path)


class StdOutWrapper(io.TextIOWrapper):
    def __init__(self, container: list = None, do_print=True):
        super().__init__(io.BytesIO(), encoding="utf-8")
        self.buff = ''
        self.ori_stout = sys.stdout
        self.container = container
        self.do_print = do_print

    def write(self, output_stream):
        if self.do_print:
            self.buff += output_stream
        if self.container is not None:
            self.container.append(output_stream)

    def flush(self):
        self.buff = ''


class StdOutLoggerWrapper:
    def __init__(self, log_file_path):
        self.ori_stout = sys.stdout
        self.log_file = open(log_file_path, "w", encoding="utf-8", buffering=1)

    def write(self, output_stream):
        self.ori_stout.write(output_stream)
        self.log_file.write(output_stream)

    def flush(self):
        self.ori_stout.flush()

    def close_file(self):
        self.log_file.close()

    def isatty(self):
        return True


def copytree(src, dst, ignore_dirs: list = None, ignore_files: list = None):
    """
    复制整个目录树
    最开始是用shutil.copytree()，但奈何Python3.7和3.8差别挺大，算了忽略这点效率吧，反正是在打包过程中，不影响用户
    :param src: 源路径
    :param dst: 目标路径
    :param ignore_dirs: 忽略的文件夹名
    :param ignore_files: 忽略的文件名
    """
    if ignore_dirs is None:
        rel_ignore_dirs = list()
    else:
        rel_ignore_dirs = list()
        for d in ignore_dirs:
            try:
                rel_ignore_dirs.append(os.path.relpath(os.path.abspath(d), src))
            except ValueError:
                continue
    if ignore_files is None:
        ignore_files = list()
    else:
        ignore_files = [os.path.abspath(os.path.join(src, f)) for f in ignore_files]
    if not os.path.exists(dst):
        os.makedirs(dst)

    copy_info = dict()
    if os.path.exists(src):
        dir_cache = "-%NONE-FLAG%-"
        for root, dirs, files in os.walk(src):
            rel_path = os.path.relpath(root, src)
            if rel_path in rel_ignore_dirs:
                dir_cache = os.path.relpath(root, src)
                continue
            if dir_cache in rel_path and rel_path.index(dir_cache) == 0:
                continue
            dst_root = os.path.join(os.path.abspath(dst),
                                    os.path.abspath(root).replace(os.path.abspath(src), "").strip("\\"))
            for file in files:
                src_file = os.path.join(root, file)
                if os.path.abspath(src_file) in ignore_files:
                    continue
                dst_file = os.path.join(dst_root, file)
                if not os.path.exists(dst_root):
                    os.makedirs(dst_root, exist_ok=True)
                copy_info[src_file] = dst_file

        progressbar = TProgressBar(msg="正在拷贝文件", max_len=len(copy_info) + 1)
        for k_id, k in enumerate(copy_info):
            shutil.copy(k, copy_info[k])
            progressbar.step()


class FileSerialize:
    def __init__(self, file_path):
        with open(file_path, "r", encoding="utf-8") as file:
            self._data = file.read()

    def get_serialize_data(self):
        return self._data

    @staticmethod
    def serialize2file(data):
        tmp_path = get_qpt_tmp_path()
        file_path = os.path.join(tmp_path, "FileSerialize.tmp")
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(data)
        return file_path


def warning_msg_box(title="Warning - GitHub: QPT-Family/QPT", text="", force=False):
    """
    发出警告框
    :param title: 标题
    :param text: 文本
    :param force: 是否强制只有确定按钮
    :return: 用户反馈
    """
    win32api = QPT_MEMORY.get_win32api
    win32con = QPT_MEMORY.get_win32con
    if force:
        flag = win32con.MB_OK | win32con.MB_ICONEXCLAMATION
    else:
        flag = win32con.MB_OKCANCEL | win32con.MB_ICONEXCLAMATION
    msg = win32api.MessageBox(0, text, title, flag)
    if not force and msg == 2:
        return False
    else:
        return True


# Author: Acer Zhang
# Datetime: 2021/6/3 
# Copyright belongs to the author.
# Please indicate the source for reprinting.
import os

from qpt.kernel.qlog import Logging, change_none_color
from qpt.executor import CreateExecutableModule

import click

# 终端运行不采用颜色方案
change_none_color()


@click.command()
@click.option("-f",
              "--folder",
              type=str,
              prompt="[项目文件夹]请输入待打包的文件夹路径",
              help='[项目文件夹]待打包的文件夹路径，该目录也应当为整个项目的根目录或工作目录，否则可能会导致出现找不到模块等Python基础报错。',
              required=True)
@click.option("-p",
              "--py",
              type=str,
              prompt="[主程序文件]请输入待打包的主要Py脚本文件路径",
              help="[主程序文件]待打包的主要Py脚本文件所在的路径，例如要yyy/xxx.py中xxx.py是需要打包的主要Python文件，那么该处填写xxx.py即可。",
              required=True)
@click.option("-s", "--save", type=str, prompt="[输出目录]请输入打包后文件保存的路径", help='[输出目录]打包后文件保存的路径。', required=True)
@click.option("-r",
              "--require",
              type=str,
              default="auto",
              prompt="[Python依赖]请输入requirements.txt文件路径，若无该文件可直接按下[回车键]自动生成该文件",
              help="[Python依赖]自动检测软件包依赖，填写auto后将会自动查找待打包的文件夹路径中所有py文件的import使用情况，最终生成requirements文件\n"
                   "当然，也可传入requirements.txt文件路径，这样即可指定依赖列表进行安装。")
@click.option("-h",
              "--hidden",
              default=False,
              type=bool,
              help="[终端窗口]是否隐藏全部终端窗口，若输入true或判定为真则隐藏全部Terminal窗口（适用于使用了PyQT、TK等桌面级可视化程序），"
                   "输入false或判定为否则不隐藏（适用于Console & 终端程序）。")
@click.option("-i",
              "--icon",
              default=None,
              type=str,
              help="[自定义图标文件]传入自定义图标文件路径，为EXE设置属于你的图标样式，支持将exe文件设置为ico/JPG/PNG等格式的自定义图标。")
def cli(folder,
        py,
        save,
        require,
        hidden,
        icon):
    Logging.info("-----------------------------QPT--------------------------------")
    Logging.info("当前执行模式为命令式执行，仅提供QPT基础功能，高阶操作可在GitHub参考最新文档")
    Logging.info("            https://github.com/GT-ZhangAcer/QPT")
    Logging.info("-----------------------------QPT--------------------------------")
    Logging.info(f"[--folder]待打包的文件夹路径为\t{os.path.abspath(folder)}")
    Logging.info(f"[--py]待打包的主Python文件路径为\t{os.path.abspath(py)}")
    Logging.info(f"[--save]打包后文件保存路径为\t{os.path.abspath(save)}")
    if require.lower() == "auto":
        Logging.info(f"[--require]使用自动化依赖查找Module：AutoRequirementsPackage")
    else:
        Logging.info(f"[--require]使依赖列表文件路径为：{os.path.abspath(require)}")
    if hidden:
        Logging.info("[--hidden]当前打包模式为：隐藏全部的Terminal窗口")
    else:
        Logging.info("[--hidden]当前打包模式为：显示Terminal窗口（就是用户打开后的那个黑色框框，可通过指定 -h True 命令来隐藏）")
    if icon:
        Logging.info(f"f[--icon]生成exe将自定义图标为该文件样式{os.path.abspath(icon)}")
    Logging.info("QPT使用命令可输入：qpt --help 获取")
    module = CreateExecutableModule(work_dir=folder,
                                    launcher_py_path=py,
                                    save_path=save,
                                    requirements_file=require,
                                    hidden_terminal=hidden,
                                    icon=icon)
    module.make()


if __name__ == '__main__':
    cli()



# Author: Acer Zhang
# Datetime:2021/7/18
# Copyright belongs to the author.
# Please indicate the source for reprinting.

import os
import ast
from typing import Optional, List, Container

# 对pip低版本做兼容
try:
    from pip._internal.utils.misc import Distribution, cast
    def get_installed_distributions(
            local_only=True,  # type: bool
            include_editables=True,  # type: bool
            editables_only=False,  # type: bool
            user_only=False,  # type: bool
            paths=None,  # type: Optional[List[str]]
    ):
        # type: (...) -> List[Distribution]
        """Return a list of installed Distribution objects.
        Left for compatibility until direct pkg_resources uses are refactored out.
        """
        from pip._internal.metadata import get_default_environment, get_environment
        from pip._internal.metadata.pkg_resources import Distribution as _Dist

        if paths is None:
            env = get_default_environment()
        else:
            env = get_environment(paths)
        dists = env.iter_installed_distributions(
            local_only=local_only,
            skip={"python", "wsgiref", "argparse"},
            include_editables=include_editables,
            editables_only=editables_only,
            user_only=user_only,
        )
        return [cast(_Dist, dist)._dist for dist in dists]
except ImportError:
    from pip._internal.utils.misc import get_installed_distributions

from qpt.memory import QPT_MEMORY, PYTHON_IGNORE_DIRS, IGNORE_PACKAGES
from qpt.kernel.qlog import TProgressBar

PACKAGE_FLAG = ".dist-info"


# 字典尽量用get，都是泪
# 尽量统一lower依赖a


class PythonPackages:
    def __init__(self):
        pass

    @staticmethod
    def search_packages_dist_info(site_package_path=None):
        """
        获取对应site_package_path下所有Python包的版本号以及TopName
        :param site_package_path:检索的路径
        :return:所有包名与包版本字典、所有包import名与包名字典、所有包与其依赖的包版本字典所构成的字典
        """
        # 获取依赖列表
        dep_pkg_dict = PythonPackages.search_dep()

        # 获取当前环境下安装的文件列表
        if site_package_path is None:
            site_package_path = QPT_MEMORY.site_packages_path
        packages_dir_list = os.listdir(site_package_path)
        packages_dist = dict()
        tops_dist = dict()
        for package_dist in packages_dir_list:
            if PACKAGE_FLAG == package_dist[-len(PACKAGE_FLAG):]:
                package = package_dist[:-len(PACKAGE_FLAG)]
                name = package[:package.rfind("-")].lower()
                # 对name修复可能的下划线情况
                if name.replace("_", "-") in dep_pkg_dict:
                    name = name.replace("_", "-")
                # 修复~的情况
                if "~" == name[:1]:
                    metadata = os.path.join(site_package_path, package_dist, "METADATA")
                    if os.path.exists(metadata):
                        with open(metadata, "r", encoding="utf-8") as metadata:
                            for metadata_line in metadata.readlines():
                                if "Name: " == metadata_line[:6]:
                                    name = metadata_line.strip("Name: ").strip("\n")
                                    break

                version = package[package.rfind("-") + 1:]
                top_file_path = os.path.join(site_package_path, package_dist, "top_level.txt")
                if os.path.exists(top_file_path):
                    with open(top_file_path, "r", encoding="utf-8") as top_file:
                        tops = top_file.readlines()

                    for top in tops:
                        # 避免路径导入
                        if "\\" in top:
                            top = top.split("\\")[-1]
                        # 兼容Linux
                        if "/" in top:
                            top = top.split("/")[-1]
                        tops_dist[top.strip("\n").lower()] = name
                packages_dist[name] = version
        return packages_dist, tops_dist, dep_pkg_dict

    @staticmethod
    def search_dep():
        """
        获取当前已安装的包以及其依赖
        :return: 所有包与其依赖的包版本字典所构成的字典
        """
        pkgs = get_installed_distributions()
        pkg_dict = dict()
        for pkg in pkgs:
            dep = pkg.requires()
            if dep:
                dep_dict = dict()
                for d in dep:
                    d_name, _, d_version = d.hashCmp[:3]
                    d_name = d_name.lower()
                    f_version = str(d_version)
                    if f_version:
                        dep_dict[d_name] = f_version
                    else:
                        dep_dict[d_name] = None
                pkg_dict[pkg.project_name.lower()] = dep_dict
            else:
                pkg_dict[pkg.project_name.lower()] = None
        return pkg_dict

    @staticmethod
    def search_import_in_text(contents):
        """
        搜索对应文本中有那些符合的Import名
        :param contents: 文本
        :return: 模块名集合
        """
        import_module = set()
        tree = ast.parse(contents)
        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for sub_node in node.names:
                    import_module.add(sub_node.name.split(".")[0])
            elif isinstance(node, ast.ImportFrom):
                if node.module:
                    import_module.add(node.module.split(".")[0])
        return import_module

    @staticmethod
    def search_import_in_dir(path, lower=True):
        """
        在对应目录中搜索所有导入的Python模块
        :param path: 路径
        :param lower: 统一小写模块搜索结果
        :return: 模块名集合
        """
        import_modules = set()
        file_path_list = list()
        for root, dirs, files in os.walk(path):
            if not (os.path.basename(root) in PYTHON_IGNORE_DIRS or "site-packages" in root):
                for file in files:
                    if os.path.splitext(file)[-1] == ".py":
                        file_path = os.path.join(root, file)
                        file_path_list.append(file_path)

        tpb = TProgressBar("正在搜索依赖", max_len=len(file_path_list))
        for file_path in file_path_list:
            with open(file_path, "r", encoding="utf-8") as f:
                import_module = PythonPackages.search_import_in_text(f.read())
                if import_module:
                    if lower:
                        import_module = set([ipm.lower() for ipm in import_module])
                    import_modules = import_modules.union(import_module)
                tpb.step(add_end_info=f"对应文件:{file_path}")
        return import_modules

    @staticmethod
    def intelligent_analysis(path, return_all_info=False):
        """
        分析对应目录下所使用的Python包情况
        :param path: 对应目录
        :param return_all_info: 是否返回所有信息，默认只返回包名与包版本字典，为True后返回包名与依赖名+版本号的字典以及被忽略的Top依赖包
        """
        # ToDo 以后加个参数，兼容非当前环境下的智能分析
        install_dict, top_dict, dep = PythonPackages.search_packages_dist_info()
        package_import = PythonPackages.search_import_in_dir(path)

        # 整合包名，避免~的情况
        for inp_name, package_name in top_dict.copy().items():
            if "~" == inp_name[:1]:
                for package in package_import:
                    if inp_name[1:] in package:
                        top_dict.pop(inp_name)
                        top_dict[package] = package_name
                        break
        for inp_name, package_name in top_dict.copy().items():
            if "~" == inp_name[:1]:
                for package in package_import:
                    if inp_name[1:] in package:
                        top_dict.pop(inp_name)
                        top_dict[package] = package_name
                        break

        # 提取显式的依赖项
        sub_requires = dict()
        for package in package_import:
            if package in top_dict and package not in IGNORE_PACKAGES:
                p_name = top_dict[package]
                if package in top_dict:
                    p_v = dep.get(p_name) if isinstance(dep.get(p_name), str) else install_dict.get(p_name)
                else:
                    p_v = install_dict.get(p_name)
                sub_requires[p_name] = p_v

        # 向顶部依赖进化
        top_deps = dict([(d_k, d_v) for d_k, d_v in dep.items() if d_v])
        requires = dict(sub_requires)
        # 遍历所有带有子依赖的节点
        for top_d in top_deps:
            # 遍历所有已经得到的显式依赖
            for sub_require in sub_requires:
                # 如果显式依赖存在子依赖中
                if sub_require in top_deps.get(top_d) and sub_require in requires:
                    # 删除该显式依赖
                    requires.pop(sub_require)
                    # 上升为主依赖并得到当前安装的版本号
                    requires[top_d] = install_dict.get(top_d)

        if return_all_info:
            # 搜索非子依赖且pip中有安装的Python包
            ignore_packages = dict()
            top_dep_flatten = set(IGNORE_PACKAGES)
            for top_dep in top_deps.values():
                top_dep_flatten = top_dep_flatten.union(top_dep)
            for install_package in install_dict:
                # 若安装的Python包不在搜索结果中，且不属于子依赖，则加入返回列表
                if install_package not in requires and install_package not in top_dep_flatten:
                    ignore_packages[install_package] = install_dict[install_package]

            return requires, dep, ignore_packages
        else:
            return requires


if __name__ == '__main__':
    _test_workdir = "./"
    print("------当前环境下包安装情况以及对应表------")
    for _i in PythonPackages.search_packages_dist_info():
        print("TEST_A", _i)
    print("------当前环境下包依赖情况------")
    for _i in PythonPackages.search_dep():
        print("TEST_B", _i)
    print("------测试场景的依赖搜索情况------")
    for _i in PythonPackages.search_import_in_dir(_test_workdir, True):
        print("TEST_C", _i)

    print("------最终结果------")
    for _i in PythonPackages.intelligent_analysis(_test_workdir):
        print(_i)



# Author: Acer Zhang
# Datetime:2021/7/3 
# Copyright belongs to the author.
# Please indicate the source for reprinting.
import os

from qpt.kernel.qlog import Logging
from qpt.kernel.qos import get_qpt_tmp_path
from qpt.modules.package import _RequirementsPackage, DEFAULT_DEPLOY_MODE
from qpt.modules.paddle_family import PaddlePaddlePackage
from qpt.memory import QPT_MEMORY


class AutoRequirementsPackage(_RequirementsPackage):
    """
    注意，这并不是个普通的Module
    """

    def __init__(self,
                 path,
                 deploy_mode=DEFAULT_DEPLOY_MODE):
        """
        自动获取Requirements
        :param path: 待扫描的文件夹路径或requirements文件路径，若提供了requirements文件路径则不会自动分析依赖情况
        :param deploy_mode: 部署模式
        """
        if not os.path.exists(path):
            Logging.info(f"当前路径{os.path.abspath(path)}中不存在Requirements文件，请优先检查路径是否提供正确，必要时使用绝对路径")
        if os.path.isfile(path):
            Logging.info(f"正在读取{os.path.abspath(path)}下的依赖情况...")
            requirements = QPT_MEMORY.pip_tool.analyze_requirements_file(path)
        else:
            Logging.info(f"[Auto]正在分析{os.path.abspath(path)}下的依赖情况...")
            requirements = QPT_MEMORY.pip_tool.analyze_dependence(path, return_path=False)

        # module_name_list = [m.name for m in module_list]
        # 对特殊包进行过滤和特殊化
        pre_add_module = list()
        for requirement in dict(requirements):
            if requirement in SPECIAL_MODULE:
                special_module, parameter = SPECIAL_MODULE[requirement]
                parameter["version"] = requirements[requirement]
                parameter["deploy_mode"] = deploy_mode
                module = special_module(**parameter)
                # # 如果开发者没有定义这个Module，那么则添加Module - ToDo 等自定义算子出了再考虑，可以在执行时候考虑
                # if module.name not in module_name_list:
                #     module_list.append(module)
                pre_add_module.append(module)
                requirements.pop(requirement)

        # 保存依赖至
        requirements_path = os.path.join(get_qpt_tmp_path(), "requirements_dev.txt")
        QPT_MEMORY.pip_tool.save_requirements_file(requirements, requirements_path)

        # 执行常规的安装
        super().__init__(requirements_file_path=requirements_path,
                         deploy_mode=deploy_mode)
        for pam in pre_add_module:
            self.add_ext_module(pam)


SPECIAL_MODULE = {"paddlepaddle": (PaddlePaddlePackage, {"version": None,
                                                         "include_cuda": False,
                                                         "deploy_mode": DEFAULT_DEPLOY_MODE}),
                  "paddlepaddle-gpu": (PaddlePaddlePackage, {"version": None,
                                                             "include_cuda": True,
                                                             "deploy_mode": DEFAULT_DEPLOY_MODE})}



import subprocess
from qpt.kernel.qlog import Logging
from qpt.memory import get_env_vars, QPT_MODE

TERMINAL_NAME = "powershell"
# TERMINAL_NAME = "pwsh"


class TerminalCallback:
    def __init__(self):
        self.cache = ""
        self.error_fitter = list()
        self.normal_fitter = list()

    def handle(self, terminal=None):
        """
        该函数将获取终端输出，并且需要显式调用执行成功与失败的func
        """
        raise NotImplementedError(f"{self.__class__.__name__}中未定义handle方法")

    def normal_func(self):
        """
        终端执行成功后需要执行的Method
        """
        raise NotImplementedError(f"{self.__class__.__name__}中未定义normal_func方法")

    def error_func(self):
        """
        终端执行失败后需要执行的Method
        """
        raise NotImplementedError(f"{self.__class__.__name__}中未定义error_func方法")


class MessageBoxTerminalCallback(TerminalCallback):
    def handle(self, terminal=None):
        pass

    def normal_func(self):
        pass

    def error_func(self):
        pass


class LoggingTerminalCallback(TerminalCallback):
    def handle(self, terminal=None):
        assert terminal, "此处需要terminal"
        line = True
        end = False
        while line:
            line = terminal.stdout.readline()
            msg = line.decode('gbk', errors="ignore").strip("b'").strip("\n").strip("\r")
            if msg:
                if msg == '---QPT OUTPUT STATUS CODE---':
                    end = True
                    continue
                if msg == "\r" or msg == "\n":
                    continue
                if end:
                    fitter_flag = False
                    for normal_fitter in self.normal_fitter:
                        if normal_fitter in self.cache:
                            self.normal_func()
                            fitter_flag = True
                    for error_fitter in self.error_fitter:
                        if error_fitter in self.cache:
                            self.error_func()
                            fitter_flag = True
                    if fitter_flag:
                        break
                    if msg == "True":
                        self.normal_func()
                        break
                    elif msg == "False":
                        self.error_func()
                        break
                self.cache += msg + "\n"
                self.print_func(msg)

    @staticmethod
    def print_func(msg):
        Logging.debug(msg)

    def normal_func(self):
        self.cache = ""
        Logging.debug("终端命令执行成功！")

    def error_func(self):
        Logging.error(f"在执行终端命令时检测到了失败，完整信息如下：\n{self.cache}")
        self.cache = ""


class RunTerminalCallback(LoggingTerminalCallback):
    @staticmethod
    def print_func(msg):
        print(msg)


class Terminal:
    """
    Terminal基类，定义Terminal基本方法
    """

    def __init__(self):
        self.main_terminal = None
        self.init_terminal()
        Logging.debug(f"正在连接{self.__class__.__name__}")

    def init_terminal(self):
        raise NotImplementedError(f"{self.__class__.__name__}中未定义init_terminal方法")

    @staticmethod
    def _get_env_vars():
        """
        为Terminal设置环境变量
        """
        # ToDO 需考虑增加兼容性支持 - 当前只考虑Windows和完整Python环境
        # ToDO 貌似没考虑打包时候的环境变量
        path_vars = get_env_vars()
        return path_vars

    def reset_terminal(self):
        """
        重启Terminal
        """
        raise NotImplementedError(f"{self.__class__.__name__}中未定义reset_terminal方法")

    def close_terminal(self):
        """
        关闭Terminal
        """
        raise NotImplementedError(f"{self.__class__.__name__}中未定义close_terminal方法")

    def shell(self, shell, callback: TerminalCallback = None):
        """
        执行shell语句
        :param shell: shell语句
        :param callback:
        """
        raise NotImplementedError(f"{self.__class__.__name__}中未定义shell方法")

    def shell_func(self, callback: TerminalCallback = None):
        """
        获取shell函数
        :param callback:
        :return: shell函数
        """
        raise NotImplementedError(f"{self.__class__.__name__}中未定义shell_func方法")


class PTerminal(Terminal):
    def __init__(self, shell_mode=True):
        self.shell_mode = shell_mode
        super(PTerminal, self).__init__()

    def init_terminal(self):
        if self.shell_mode:
            self.main_terminal = subprocess.Popen(TERMINAL_NAME,
                                                  stdout=subprocess.PIPE,
                                                  stderr=subprocess.STDOUT,
                                                  stdin=subprocess.PIPE,
                                                  shell=True,
                                                  env=self._get_env_vars())
        else:
            self.main_terminal = subprocess.Popen(TERMINAL_NAME,
                                                  shell=False,
                                                  env=self._get_env_vars())
        # ToDo 加个自动判断
        Logging.info("\n如在本信息之后停留时间较长，请升级Windows Powershell至版本5即可解决该问题，下载地址：\n"
                     "官方地址：https://www.microsoft.com/en-us/download/details.aspx?id=54616\n"
                     "团子云镜像：https://s.dango.cloud/s/ZyDSn 下载码：egl95d")
        prepare = "chcp 65001"
        self._shell_func()(prepare)
        if QPT_MODE == "Debug":
            # 打印环境变量
            self._shell_func()("dir env:")
        # prepare = "$env:PATH='" + self._get_env_vars() + "'"
        # self._shell_func()(prepare)

    def reset_terminal(self):
        self.close_terminal()
        self.init_terminal()

    def close_terminal(self):
        self.main_terminal.terminate()

    def _shell_func(self, callback: TerminalCallback = LoggingTerminalCallback()):
        # ToDo 实现Callback
        def closure(closure_shell):
            Logging.debug(f"SHELL: {closure_shell}")
            if closure_shell[1:3] == ":\\":
                closure_shell += f"cd {closure_shell[:2]} ;"
            closure_shell = f'{closure_shell} ; echo "---QPT OUTPUT STATUS CODE---" $? \n'
            try:
                final_shell = closure_shell.encode("utf-8")
            except Exception as e:
                Logging.error("执行该指令时遇到解码问题，目前将采用兼容模式进行，原始报错如下：\n" + str(e))
                final_shell = closure_shell.encode("utf-8", errors="ignore")
            self.main_terminal.stdin.write(final_shell)
            try:
                self.main_terminal.stdin.flush()
            except Exception as e:
                Logging.error(str(e))
            callback.handle(self.main_terminal)

        return closure

    def shell_func(self, callback: TerminalCallback = None):
        return self._shell_func(callback)

    def shell(self, shell, callback: TerminalCallback = LoggingTerminalCallback()):
        self.shell_func(callback)(shell)


if __name__ == '__main__':
    t = PTerminal()
    t.shell("ping 192.168.1.1")



# Author: Acer Zhang
# Datetime:2021/7/3 
# Copyright belongs to the author.
# Please indicate the source for reprinting.

import platform
import os
from distutils.sysconfig import get_python_lib

from qpt.kernel.qlog import Logging


def init_wrapper(func):
    @property
    def render(self):
        if func.__name__ in self.memory:
            out = self.memory[func.__name__]
        else:
            out = func(self)
            self.memory[func.__name__] = out
        return out

    return render


class QPTMemory:
    def __init__(self):
        self.memory = dict()

    def set_mem(self, name, variable):
        self.memory[name] = variable
        return variable

    def free_mem(self, name):
        self.memory.pop(name)

    @init_wrapper
    def platform_bit(self):
        arc = platform.machine()
        Logging.debug(f"操作系统位数：{arc}")
        return arc

    @init_wrapper
    def platform_os(self):
        p_os = platform.system()
        Logging.debug(f"操作系统类型：{p_os}")
        return p_os

    @init_wrapper
    def site_packages_path(self):
        site_package_path = os.path.abspath(get_python_lib())
        return site_package_path

    @init_wrapper
    def pip_tool(self):
        from qpt.kernel.qinterpreter import PipTools
        pip_tools = PipTools()
        return pip_tools

    @init_wrapper
    def get_win32con(self):
        import win32con
        return win32con

    @init_wrapper
    def get_win32api(self):
        import win32api
        return win32api


QPT_MEMORY = QPTMemory()


def check_bit():
    arc = QPT_MEMORY.platform_bit
    assert "64" in arc, "当前QPT不支持32位操作系统"


def check_os():
    p_os = QPT_MEMORY.platform_os
    assert "Windows" in p_os, "当前QPT只支持Windows系统"


IGNORE_ENV_FIELD = ["conda", "Conda", "Python", "python"]


def get_env_vars(work_dir="."):
    """
    获取当前待设置的环境变量字典
    :param work_dir:
    :return: dict
    """
    env_vars = dict()
    # Set PATH ENV
    path_env = os.environ.get("PATH").split(";")
    pre_add_env = os.path.abspath("./Python/Lib/site-packages") + ";" + \
                  os.path.abspath("./Python/Lib") + ";" + \
                  os.path.abspath("./Python/Lib/ext") + ";" + \
                  os.path.abspath("./Python") + ";" + \
                  os.path.abspath("./Python/Scripts") + ";"

    for pe in path_env:
        if pe:
            add_flag = True
            for ief in IGNORE_ENV_FIELD:
                if ief in pe:
                    add_flag = False
                    break
            if add_flag:
                pre_add_env += pe + ";"
    env_vars["PATH"] = pre_add_env

    # Set PYTHON PATH ENV
    env_vars["PYTHONPATH"] = os.path.abspath("./Python/Lib/site-packages") + ";" + \
                             work_dir + ";" + \
                             os.path.abspath("./Python")
    os_env = os.environ.copy()
    os_env.update(env_vars)

    return os_env


PYTHON_IGNORE_DIRS = [".idea", ".git", ".github", "venv"]

# 被忽略的Python包
IGNORE_PACKAGES = ["virtualenv", "pip", "setuptools", "cpython"]

# QPT运行状态 Run/Debug
QPT_MODE = os.getenv("QPT_MODE")

# QPT检测到的运行状态 Run/本地Run
QPT_RUN_MODE = None


class CheckRun:
    @staticmethod
    def make_run_file(configs_path):
        with open(os.path.join(configs_path, "run_act.lock"), "w") as f:
            f.write("Run Done")

    @staticmethod
    def check_run_file(configs_path):
        global QPT_RUN_MODE
        if QPT_RUN_MODE is None:
            QPT_RUN_MODE = os.path.exists(os.path.join(configs_path, "run_act.lock"))
        return QPT_RUN_MODE


def check_all():
    # 检查系统
    check_os()
    # 检查arc
    check_bit()


check_all()



# Author: Acer Zhang
# Datetime: 2021/7/12
# Copyright belongs to the author.
# Please indicate the source for reprinting.

import os
import sys

from qpt.kernel.qos import copytree
from qpt.modules.base import SubModule, SubModuleOpt


class SearchCUDA:
    def __init__(self):
        self.paths = dict()

    def search_sys(self):
        for a_v in range(9, 13):
            for b_v in range(5):
                path = os.environ.get(f"CUDA_PATH_V{a_v}_{b_v}")
                if path:
                    bin_path = os.path.join(path, "bin")
                    self.paths[str(a_v) + "." + str(b_v)] = bin_path

    def search_conda(self):
        pass

    def get_paths(self):
        return self.paths


class CopyCUDADLL(SubModuleOpt):
    def __init__(self, cuda_version):
        super(CopyCUDADLL, self).__init__()
        self.cuda_version = cuda_version

    def act(self) -> None:
        if self.cuda_version is None:
            raise Exception("暂不接受不指定版本的情况")
        else:
            version = self.cuda_version.split(".")
            assert len(version) == 2, "CUDA版本号需要为以下格式传入：主版本号.从版本号，例如11.0"
            base_path = os.environ.get(f"CUDA_PATH_V{version[0]}_{version[1]}")
            if not base_path or not os.path.exists(base_path):
                raise FileNotFoundError(f"当前环境的{os.path.abspath(base_path)}目录下无CUDA驱动，无法封装。")
            bin_path = os.path.join(base_path, "bin")
            copytree(bin_path, os.path.join(self.module_path, "opt/CUDA"), ignore_files=["compute-sanitizer.bat"])


class SetCUDAEnv(SubModuleOpt):
    def __init__(self):
        super(SetCUDAEnv, self).__init__()

    def act(self) -> None:
        cuda_path = os.path.join(os.path.abspath(self.module_path), "opt/CUDA")
        os.environ["PATH"] += ";" + cuda_path
        sys.path.append(cuda_path)


class CopyCUDAPackage(SubModule):
    def __init__(self, cuda_version):
        super().__init__()
        self.add_pack_opt(CopyCUDADLL(cuda_version=cuda_version))
        self.add_unpack_opt(SetCUDAEnv())



# Author: Acer Zhang
# Datetime:2021/10/8 
# Copyright belongs to the author.
# Please indicate the source for reprinting.
import threading
import traceback
import gc

import tkinter
import tkinter.font
from tkinter import ttk
from tkinter.messagebox import showerror


class ProgressbarFrame:
    def __init__(self,
                 bind_fuc,
                 max_step: int,
                 default_text: str = "正在准备",
                 title: str = "初始化程序 - QPT",
                 icon=None):
        from ttkbootstrap import Style

        self.scale = 1. / (max_step + 1)
        self.count = 0
        # 窗体构建
        self.root = tkinter.Tk()
        self.root.geometry("666x25")
        self.root.style = Style("lumen")
        self.root.title(title)
        self.root.wm_resizable(False, False)
        self.root.iconbitmap(icon)
        default_font = tkinter.font.nametofont("TkDefaultFont")
        # ToDo 此处需考虑字体在不同操作系统兼容性
        default_font.configure(family="等线", size=10)
        self.root.option_add("*Font", "TkDefaultFont")

        self.progressbar_var = tkinter.IntVar(self.root, value=0)
        self.label_var = tkinter.StringVar(self.root, value=default_text)
        self.value_var = tkinter.StringVar(self.root, value="进度" + f"{0:.2f} %".rjust(10, " "))
        self.progressbar_var.trace("w", self.progressbar_var_trace)

        label = ttk.Label(self.root,
                          textvariable=self.label_var,
                          style="TLabel",
                          width=20)
        label.pack(side="left", padx=5)

        progressbar = ttk.Progressbar(self.root,
                                      variable=self.progressbar_var,
                                      style="Striped.Horizontal.TProgressbar")
        progressbar.pack(side="left", fill="x", expand="yes", padx=5, pady=2)

        value = ttk.Label(self.root,
                          textvariable=self.value_var,
                          style="TLabel",
                          width=12)
        value.pack(side="left", fill="x", padx=1)

        self.thread = None

        def render():
            def func(self):
                try:
                    bind_fuc(self)
                    self.step("加载完毕")
                except Exception as e:
                    msg = traceback.format_exc()
                    showerror(title="发生异常 - QPT提示", message=f"简略异常说明:\n{e}\n\n完整报错信息如下：\n{msg}")
                finally:
                    self.close()
                    gc.collect()
                return func

            self.thread = threading.Thread(target=func, args=(self,))
            self.thread.setDaemon(True)
            self.thread.start()

        self.root.after(ms=32, func=render)

        self.root.mainloop()

    def progressbar_var_trace(self, *args):
        v = self.progressbar_var.get()
        self.value_var.set(f"进度 {v:.2f} %")

    def close(self):
        self.root.destroy()
        # self.root.quit()

    def step(self, text: str = None):
        if text:
            self.label_var.set(text)
        self.progressbar_var.set(self.scale * (self.count + 1) * 100)
        self.count += 1


if __name__ == '__main__':
    import time


    def foo(self):
        for i in range(10):
            print(2)
            time.sleep(0.2)
            self.step()
        self.close()


    ProgressbarFrame(foo, 10)



# Author: Acer Zhang
# Datetime:2021/6/24 
# Copyright belongs to the author.
# Please indicate the source for reprinting.

# 设置镜像源
import os

from qpt import Logging
from qpt.kernel.qinterpreter import PIPTerminal, PipTools
from qpt.memory import QPT_MEMORY


def set_default_pip_source(source: str):
    QPT_MEMORY.pip_tool.source = source
    Logging.debug(f"已设置PIP镜像源为：{source}")


def set_default_pip_lib(interpreter_path: str):
    if not os.path.splitext(interpreter_path)[1]:
        interpreter_path = os.path.join(interpreter_path, "python.exe")
    QPT_MEMORY.pip_tool.pip_main = PIPTerminal(interpreter_path).shell_func()
    Logging.debug(f"已设置PIP跨版本编译模式，目标解释器路径为：{interpreter_path}")


def set_pip_configs(lib_package_path=None,
                    source: str = None):
    """
    设置全局pip工具组件
    :param lib_package_path: pip所在位置
    :param source: 镜像源地址
    """
    QPT_MEMORY.set_mem("pip_tool", PipTools(pip_path=lib_package_path, source=source))



# Author: Acer Zhang
# Datetime:2021/7/3 
# Copyright belongs to the author.
# Please indicate the source for reprinting.

import os
import sys

from qpt.kernel.qlog import Logging
from qpt.modules.base import SubModule, SubModuleOpt, GENERAL_LEVEL_REDUCE, LOW_LEVEL_REDUCE
from qpt.modules.package import CustomPackage, DEFAULT_DEPLOY_MODE
from qpt.modules.cuda import CopyCUDAPackage
from qpt.memory import QPT_MEMORY


class SetPaddleFamilyEnvValueOpt(SubModuleOpt):
    def __init__(self):
        super(SetPaddleFamilyEnvValueOpt, self).__init__()

    def act(self) -> None:
        os.environ["HUB_HOME"] = os.path.join(self.module_path, "opt/HUB_HOME")
        os.environ["PPNLP_HOME"] = os.path.join(self.module_path, "opt/PPNLP_HOME")
        os.environ["SEG_HOME"] = os.path.join(self.module_path, "opt/SEG_HOME")


class CheckAVXOpt(SubModuleOpt):
    def __init__(self, version, use_cuda=False):
        super(CheckAVXOpt, self).__init__(disposable=True)
        self.version = version
        # ToDo 做CUDA的适配 + 去掉>=
        self.use_cuda = use_cuda

    @staticmethod
    def _check_dll():
        from qpt.modules.tools.check_paddle_noavx import SUPPORT_AVX
        return SUPPORT_AVX

    def act(self) -> None:
        if not self._check_dll():
            Logging.warning("为保证可以成功在NoAVX平台执行PaddlePaddle，即将忽略小版本号进行安装PaddlePaddle-NoAVX")
            Logging.warning("当前CPU不支持AVX指令集，正在尝试在线下载noavx版本的PaddlePaddle")
            QPT_MEMORY.pip_tool.pip_shell("uninstall paddlepaddle")
            if self.version:
                new_v = self.version[:self.version.rindex(".")]
                QPT_MEMORY.pip_tool.pip_shell(
                    f"install paddlepaddle=={new_v} -f https://www.paddlepaddle.org.cn/whl/mkl/stable/noavx.html"
                    " --no-index --no-deps --force-reinstall")
            else:
                QPT_MEMORY.pip_tool.pip_shell(
                    f"install paddlepaddle -f https://www.paddlepaddle.org.cn/whl/mkl/stable/noavx.html"
                    " --no-index --no-deps --force-reinstall")


def split_paddle_version(package_dist):
    """
    paddle_dist信息分割
    :param package_dist: 形如paddlepaddle_gpu-xxx.postyyy.dist-info的版本信息
    :return: xxx和yyy
    """
    package_dist = package_dist.strip(".dist-info").strip("paddlepaddle_gpu-")
    if ".post" in package_dist:
        paddle_version, cuda_version = package_dist.split(".post")
        cuda_version_a, cuda_version_b = cuda_version[:-1], cuda_version[-1]
    else:
        paddle_version = package_dist
        cuda_version_a, cuda_version_b = "10", "2"
    return paddle_version, cuda_version_a + "." + cuda_version_b


def search_paddle_cuda_version(package_dist=None):
    """
    paddle_dist信息分割，若未提供package_dist则自动搜索
    :param package_dist: 形如paddlepaddle_gpu-xxx.postyyy.dist-info的版本信息
    :return: xxx和yyy
    """
    if package_dist:
        return split_paddle_version(package_dist)
    else:
        import pip
        site_packages_path = os.path.dirname(os.path.dirname(pip.__file__))
        packages_list = os.listdir(site_packages_path)
        paddle_version, cuda_version = None, None
        for package in packages_list:
            if "paddlepaddle_gpu" in package and ".dist-info" in package:
                paddle_version, cuda_version = split_paddle_version(package)
                Logging.info(f"搜索到PaddlePaddle-GPU版本信息：{paddle_version}，所需CUDA版本：{cuda_version}")
                return paddle_version, cuda_version
        if paddle_version is None:
            Logging.info(f"当前Python解释器路径为：{sys.executable}")
            Logging.error(f"当前Python环境下没有PaddlePaddle程序包，请切换至含有PaddlePaddle-GPU的程序包下使用QPT")
            exit(2)


class PaddlePaddleCheckAVX(SubModule):
    """
    解决AVX的适配，并且给予更低优先级
    """

    def __init__(self, version, use_cuda=False):
        super(PaddlePaddleCheckAVX, self).__init__(level=LOW_LEVEL_REDUCE)
        self.add_unpack_opt(CheckAVXOpt(version=version, use_cuda=use_cuda))


class PaddlePaddlePackage(CustomPackage):
    def __init__(self,
                 version: str = None,
                 include_cuda=False,
                 deploy_mode=DEFAULT_DEPLOY_MODE):
        self.level = GENERAL_LEVEL_REDUCE
        opts = None
        if not include_cuda:
            super().__init__("paddlepaddle",
                             version=version,
                             deploy_mode=deploy_mode,
                             opts=opts)
        else:
            if version:
                paddle_version, cuda_version = search_paddle_cuda_version(version)
            else:
                Logging.flush()
                Logging.info("Requirements文件中并未指定paddlepaddle-gpu所对应的CUDA版本，请输入例如10.0、11.2 的"
                             "CUDA版本")
                cuda_version = input()
                paddle_version, _ = search_paddle_cuda_version()

            if cuda_version == "10.2":
                pass
            else:
                paddle_version += ".post" + cuda_version.replace(".", "")

            super().__init__("paddlepaddle-gpu",
                             version=paddle_version,
                             deploy_mode=deploy_mode,
                             find_links="https://paddlepaddle.org.cn/whl/mkl/stable.html")
            self.add_ext_module(CopyCUDAPackage(cuda_version=cuda_version))
        # ToDO 当前方案需要放置在init后
        self.add_unpack_opt(SetPaddleFamilyEnvValueOpt())
        self.add_ext_module(PaddlePaddleCheckAVX(version=version, use_cuda=include_cuda))


class PaddleHubPackage(CustomPackage):
    def __init__(self,
                 version: str = None,
                 deploy_mode=DEFAULT_DEPLOY_MODE):
        super().__init__("paddlehub",
                         version=version,
                         deploy_mode=deploy_mode)


class PaddleDetectionPackage(CustomPackage):
    def __init__(self,
                 version: str = None,
                 deploy_mode=DEFAULT_DEPLOY_MODE):
        super().__init__("paddledetection",
                         version=version,
                         deploy_mode=deploy_mode)


class PaddleSegPackage(CustomPackage):
    def __init__(self,
                 version: str = None,
                 deploy_mode=DEFAULT_DEPLOY_MODE):
        super().__init__("paddleseg",
                         version=version,
                         deploy_mode=deploy_mode)


class PaddleXPackage(CustomPackage):
    def __init__(self,
                 version: str = None,
                 deploy_mode=DEFAULT_DEPLOY_MODE):
        super().__init__("paddlex",
                         version=version,
                         deploy_mode=deploy_mode)


class PaddleGANPackage(CustomPackage):
    def __init__(self,
                 version: str = None,
                 deploy_mode=DEFAULT_DEPLOY_MODE):
        super().__init__("paddlegan",
                         version=version,
                         deploy_mode=deploy_mode)



import os
import sys


from qpt.memory import get_env_vars

IGNORE_ENV_FIELD = ["conda", "Conda", "Python", "python"]
interpreter_dir = os.path.dirname(sys.executable)
sys_p = sys.path
# ToDo 用户运行时可能sys.path还是旧的
new_sys_p = list()
for sp in sys_p:
    for field in IGNORE_ENV_FIELD:
        if field in sp and interpreter_dir not in sp:
            break
    else:
        new_sys_p.append(sp)

sys.path = new_sys_p
print(sys.path)
ROOT_PATH = os.path.abspath("./")
os.environ.update(get_env_vars(ROOT_PATH))
from qpt.executor import RunExecutableModule

module = RunExecutableModule("./")
module.run()



from setuptools import setup
from setuptools import find_packages
from qpt.version import version

# python setup.py sdist bdist_wheel
setup(
    name='QPT',
    version=version,
    packages=find_packages(),
    url='https://github.com/GT-ZhangAcer/QPT',
    license='LGPL',
    author='GT-ZhangAcer',
    author_email='zhangacer@foxmail.com',
    description='QPT-基于Python的快捷环境封装工具',
    install_requires=['wget',
                      "click",
                      "pefile",
                      "pillow",
                      "ttkbootstrap"],
    python_requires='>3.5',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'qpt = qpt.command:cli',
        ]}
)
